from cachetools import LRUCache
from enum import Enum
from typing import List, Union, Optional, Tuple, Dict, NamedTuple, Set
from netaddr import IPAddress, IPNetwork

from cyst.api.environment.environment import EnvironmentMessaging
from cyst.api.environment.message import StatusValue, StatusOrigin, MessageType, Status
from cyst.api.host.service import ActiveService
from cyst.api.network.elements import Connection, Route
from cyst.api.network.node import Node
from cyst.api.network.firewall import FirewallPolicy, FirewallRule, FirewallChainType
from cyst.api.utils.duration import msecs

from cyst.platform.environment.message import MessageImpl, RequestImpl, ResponseImpl
from cyst.platform.network.node import NodeImpl
from cyst.platform.network.elements import PortImpl, InterfaceImpl, Endpoint, Resolver
from cyst.platform.network.firewall import FirewallImpl

# TODO: The following is needed to automatically create a passive service representing the router itself for the purpose
#       of exploitation. This ought to be changed in the near future by transforming the router into traffic processor.
from cyst.api.logic.access import AccessLevel
from cyst.platform.host.service import PassiveServiceImpl, ActiveService


class Router(NodeImpl):

    def __init__(self, id: str, env: EnvironmentMessaging) -> None:

        super(Router, self).__init__(id, "Router")

        self._env: EnvironmentMessaging = env
        self._ports: List[PortImpl] = []
        self._local_ips: Dict[IPAddress, int] = {}
        self._local_nets: List[IPNetwork] = []
        self._routes: List[Route] = []
        self._router_ips: Set[IPAddress] = set()
        # Cache storing last 64 requests
        self._request_cache: LRUCache = LRUCache(64)

        self._fw: Optional[FirewallImpl] = None

        # create and add a passive service representing a router itself
        # WARNING: No access policy is created, so the only way to abuse this service is to perform an exploitation
        service = PassiveServiceImpl("router", "router", "1.2.3", False, AccessLevel.LIMITED)
        self.add_service(service)

        self._default_processing_time = msecs(20).to_float()

    # jsonpickle struggles with IPAddress and IPNetwork in a router, even though it has specific handlers for it,
    # so we convert it to a string and back
    def __getstate__(self) -> dict:
        result = self.__dict__
        replacement_local_ips = {}
        for ip, index in self._local_ips.items():
            replacement_local_ips[str(ip)] = index
        result["_local_ips"] = replacement_local_ips

        replacement_router_ips = set()
        for ip in self._router_ips:
            replacement_router_ips.add(str(ip))
        result["_router_ips"] = replacement_router_ips

        replacement_local_nets = []
        for net in self._local_nets:
            replacement_local_nets.append(str(net))
        result["_local_nets"] = replacement_local_nets

        return result

    def __setstate__(self, state: dict) -> None:
        original_local_ips = {}
        for ip, index in state["_local_ips"].items():
            original_local_ips[IPAddress(ip)] = index
        state["_local_ips"] = original_local_ips

        original_router_ips = set()
        for ip in state["_router_ips"]:
            original_router_ips.add(IPAddress(ip))
        state["_router_ips"] = original_router_ips

        original_local_nets = []
        for net in state["_local_nets"]:
            original_local_nets.append(IPNetwork(net))
        state["_local_nets"] = original_local_nets

        self.__dict__.update(state)

    # Override adding of traffic processor to register firewall for routing
    def add_traffic_processor(self, value: ActiveService) -> None:
        self._traffic_processors.append(value)
        if isinstance(value, FirewallImpl):
            for ip in self._router_ips:
                value.add_local_ip(ip)
            self._fw = value

    # MYPY: Inheritance issue, parent has Interface, this one PortImpl, both inherit from Port.
    # Maybe create Interface(Port) as it should be its extension or add separate ports property to Node?
    @property
    def interfaces(self) -> List[PortImpl]:
        return self._ports

    # Port removal not supported, but they can be overwritten by using already used port index
    # Otherwise the port index is just incremented
    def add_port(self, ip: Union[str, IPAddress] = "", mask: str = "", index: int = -1) -> int:
        new_index = index
        if index == -1:
            new_index = len(self._ports)

        if isinstance(ip, str):
            # Empty port configuration
            if ip == "":
                self._ports.append(PortImpl(index=new_index))
                return new_index

            ip = IPAddress(ip)

        self._ports.append(PortImpl(ip, mask, new_index))
        self._router_ips.add(ip)
        if self._fw:
            self._fw.add_local_ip(ip)
        return new_index

    def port_net(self, index: int) -> Optional[IPNetwork]:
        return self._ports[index].net

    def add_route(self, route: Route) -> None:
        self._routes.append(route)
        self._routes.sort()

    def list_routes(self) -> List[Route]:
        return self._routes

    # Check if IP belongs to local net
    def _is_local_ip(self, ip: IPAddress) -> bool:
        for net in self._local_nets:
            if ip in net:
                return True
        return False

    def _connect_node(self, node: NodeImpl, connection: Connection, router_index: int = -1, node_index: int = 0, net: str = "") -> Tuple[bool, Union[str, Tuple[int, int]]]:
        # If both a specific router index and network designation si provided, bail out
        if router_index != -1 and net:
            return False, "Cannot specify both router index and network designation"

        # No router port was selected, new one is dynamically added
        new_router_index = router_index
        if router_index == -1:
            # Net designation was provided, use it for the port
            if net:
                network = IPNetwork(net)
                new_router_index = self.add_port(next(network.iter_hosts()), str(network.netmask))
            # otherwise add and unconfigured port
            else:
                new_router_index = self.add_port()

        new_node_index = node_index
        router_port = self._ports[new_router_index]
        if new_node_index != -1:
            node_interface = node.interfaces[new_node_index] if node.interfaces else None
        else:
            node_interface = None

        # Get DHCP status
        dhcp = False if node_interface and node_interface.ip else True

        assigned_ip = None

        if dhcp:
            if new_router_index == -1 or not router_port.net:
                return False, "Trying to connect a node to a router port that does not support automatic address assignment"

            # Find a suitable host from router's network
            for h in router_port.net.iter_hosts():
                if h != router_port.ip and h not in self._local_ips:
                    assigned_ip = h
                    break

            if not assigned_ip:
                return False, "Do not have any more addresses to allocate in the range {}".format(str(router_port.net))

            node_interface = InterfaceImpl(assigned_ip, str(router_port.net.netmask))
            new_node_index = node.add_interface(node_interface)

        else:
            # If the router does not have network configured, accept the one by the connected node
            if not router_port.net:
                router_port.set_ip(node_interface.gateway)
                router_port.set_net(node_interface.net)

                self._router_ips.add(node_interface.gateway)
                if self._fw:
                    self._fw.add_local_ip(node_interface.gateway)

            # Check if there is a conflict between router's IP and expected IP from connected node
            if node_interface.gateway != router_port.ip:
                return False, "The connected node expects gateway to have an IP {}, but it has an IP {}".format(str(node_interface.gateway), str(router_port.ip))

            assigned_ip = node_interface.ip

        # Add the host ip to the list of local ips and the port net to the list of local networks
        self._local_ips[assigned_ip] = router_port.index
        if router_port.net not in self._local_nets:
            self._local_nets.append(router_port.net)

        # Set endpoints on both ends
        router_port.connect_endpoint(Endpoint(node.id, new_node_index, node_interface.ip), connection)
        node_interface.connect_gateway(router_port.ip, connection, self.id, new_router_index)

        return True, (new_router_index, new_node_index)

    def _connect_router(self, router: 'Router', connection: Connection, remote_port_index: int = -1, local_port_index: int = -1) -> Tuple[bool, Union[str, Tuple[int, int]]]:

        # Create missing ports if needed
        remote_port = remote_port_index
        local_port = local_port_index

        if remote_port == -1:
            remote_port = router.add_port()

        if local_port == -1:
            local_port = self.add_port()

        self._ports[local_port].connect_endpoint(Endpoint(router.id, remote_port), connection)
        router._ports[remote_port].connect_endpoint(Endpoint(self.id, local_port), connection)

        # After routers' connection, routes must be added manually

        return True, (remote_port, local_port)

    def add_routing_rule(self, rule: FirewallRule) -> None:
        # This can be brittle, but we assume that FW is always preprocessor no. 1
        self._fw.add_rule(FirewallChainType.FORWARD, rule)

    def remove_routing_rule(self, index) -> None:
        self._fw.remove_rule(FirewallChainType.FORWARD, index)

    def list_routing_rules(self) -> List[Tuple[FirewallChainType, FirewallPolicy, List[FirewallRule]]]:
        return self._fw.list_rules(FirewallChainType.FORWARD)

    def set_default_routing_policy(self, policy: FirewallPolicy) -> None:
        self._fw.set_default_policy(FirewallChainType.FORWARD, policy)

    def get_default_routing_policy(self) -> FirewallPolicy:
        return self._fw.get_default_policy(FirewallChainType.FORWARD)

    def routes(self, src: IPAddress, dst: IPAddress, service: str) -> Optional[PortImpl]:
        # I lost the code, so this may or may not be enough here
        # It is a butchered version of process message
        port = self._local_ips.get(dst, -1)
        if port != -1:
            if src in self._ports[port].net or self._fw.evaluate(src, dst, service)[0]:
                return self._ports[port]
            else:
                return None

        # TODO: i do not understand this check
        elif self._is_local_ip(dst):
            return None

        else:
            for route in self._routes:
                if dst in route.net:
                    return self._ports[route.port]

        return None

    async def process_message(self, message: MessageImpl, delay: float) -> Tuple[bool, float]:
        processing_delay = self._default_processing_time

        # Do not process messages that are going on for far too long
        if message.decrease_ttl() == 0:
            m = ResponseImpl(message, status=Status(StatusOrigin.NETWORK, StatusValue.FAILURE), content="TTL expired", session=message.session)
            m.set_next_hop()
            self._env.send_message(m, delay + self._default_processing_time)
            return False, processing_delay

        # Messages that are not dropped due to TTL are first inspected by the traffic processors that determine, whether
        # anything should be done with the message at all. Firewall gets a separate processing later.
        for processor in self._traffic_processors:
            if not isinstance(processor, FirewallImpl):
                result, processor_delay = await processor.process_message(message)
                processing_delay += processor_delay
                if not result:
                    return False, processing_delay

        # If message is still going through a session then pass it along where it should go...
        if message.in_session:
            message.set_next_hop()
            return True, processing_delay

        # ...the same goes for responses, which travel back from whence they came
        if message.type == MessageType.RESPONSE:
            port = self._request_cache.get(message.id, -1)
            if port != -1:
                message.set_next_hop(Endpoint(self.id, port, self._ports[port].ip), self._ports[port].endpoint)
                return True, processing_delay

        # Unless the request vanished from cache - then we have to try to deliver it the old-fashioned way

        # Check for messages running around in circles
        if message.type == MessageType.REQUEST:
            port = self._request_cache.get(message.id, -1)
            if port != -1:
                m = ResponseImpl(message, status=Status(StatusOrigin.NETWORK, StatusValue.FAILURE),
                                 content="Message stuck in a cycle", session=message.session)
                # The next hop is automatically calculated because it is a response
                m.set_next_hop(Endpoint(self.id, port, self._ports[port].ip), self._ports[port].endpoint)
                self._env.send_message(m, delay + processing_delay)
                return False, processing_delay

        # TODO evaluate permeability between networks!
        # When looking at the current target, the router must also check, if the target is within the same network as
        # is the net of the arriving port
        # The same goes for router's constituency

        # The rule of thumb is - you can cross from local networks to remote networks, but you can't cross between
        # local networks and you can't go from remote network to local networks
        # Port forwarding is in the current state impossible and is ignored to reduce scope
        # Addendum (until more robust solution is found) - the aforementioned holds, but in addition, the FW is
        # consulted to check for explicitly allowed paths for crossing between local networks

        # Check if the target is linked to a router port
        port = self._local_ips.get(message.dst_ip, -1)
        if port != -1:
            # It is and (it is in the same network or have explicit permission in the firewall/routing policy)
            # This would break when the message traversed more routers
            if self._fw.evaluate(message.non_session_path[0].src.ip, message.dst_ip, message.dst_service)[0]:
                src_ip = self._ports[port].ip
                message.set_next_hop(Endpoint(self.id, port, src_ip), self._ports[port].endpoint)
                # Store the info about incoming port to enable pass-through of responses
                if message.type == MessageType.REQUEST:
                    self._request_cache[message.id] = message.current.port
                return True, processing_delay
            # It is, but in another network
            else:
                m = ResponseImpl(message, status=Status(StatusOrigin.NETWORK, StatusValue.FAILURE),
                             content="Host unreachable", session=message.session)
                # The next hop is automatically calculated because it is a response
                m.set_next_hop()
                self._env.send_message(m, delay + self._default_processing_time)
                return False, processing_delay

        # It is not, but belongs to router's constituency
        elif self._is_local_ip(message.dst_ip):
            m = ResponseImpl(message, status=Status(StatusOrigin.NETWORK, StatusValue.FAILURE), content="Host unreachable", session=message.session)
            # The next hop is automatically calculated because it is a response
            m.set_next_hop()
            self._env.send_message(m, delay + self._default_processing_time)
            return False, processing_delay
        # Try to send it somewhere
        else:
            for route in self._routes:
                if message.dst_ip in route.net:
                    message.set_next_hop(Endpoint(self.id, route.port, self._ports[route.port].ip), self._ports[route.port].endpoint)
                    # Store the info about incoming port to enable pass-through of responses
                    if message.type == MessageType.REQUEST:
                        self._request_cache[message.id] = message.current.port
                    return True, processing_delay

            m = ResponseImpl(message, status=Status(StatusOrigin.NETWORK, StatusValue.FAILURE), content="Network address {} not routable".format(message.dst_ip), session=message.session)
            m.set_next_hop()
            self._env.send_message(m, delay + self._default_processing_time)
            return False, processing_delay

    @staticmethod
    def cast_from(o: Node) -> 'Router':
        if isinstance(o, Router):
            return o
        else:
            raise ValueError("Malformed underlying object passed with the Node interface")



from cyst.api.environment.environment import Environment
from cyst.api.configuration import NodeConfig, PassiveServiceConfig, AccessLevel

# ----------------------------------------------------------------------------------------------------------------------
# Addition to 01_do_nothing.py
# ----------------------------------------------------------------------------------------------------------------------
target = NodeConfig(
    active_services=[],
    passive_services=[
        PassiveServiceConfig(
            type="bash",
            owner="root",
            version="8.1.0",
            access_level=AccessLevel.LIMITED,
            local=True,
            id="bash_service"
        )
    ],
    shell="bash",
    interfaces=[],
    traffic_processors=[],
    id="target"
)

e = Environment.create().configure(target)
# ----------------------------------------------------------------------------------------------------------------------
e.control.init()
e.control.run()
e.control.commit()

stats = e.infrastructure.statistics
print(f"Run id: {stats.run_id}\nStart time real: {stats.start_time_real}\n"
      f"End time real: {stats.end_time_real}\nDuration virtual: {stats.end_time_virtual}")



from dataclasses import dataclass, field
from flags import Flags
from typing import Optional, Dict, Any, List, Union, Tuple
from uuid import uuid4

from serde import serialize, coerce

from cyst.api.configuration.configuration import ConfigItem
from cyst.api.configuration.logic.access import AuthorizationConfig, AccessSchemeConfig, AuthenticationProviderConfig
from cyst.api.configuration.logic.data import DataConfig

from cyst.api.logic.access import AccessLevel


class ServiceParameter(Flags):
    """
    Service parameter represents a domain of parametrization for passive services.

    Values
        :ENABLE_SESSION: A service can be a destination of a session, e.g., SSH or VPN tunnel, or HTTP server.
            Possible values: True|False
        :SESSION_ACCESS_LEVEL: An access level of a session when it is established. This can be different from the
            service access level, e.g., SSH daemon has an elevated service access level, but its sessions are always
            limited. Possible values: the domain of cyst.api.logic.access.AccessLevel.
    """
    ENABLE_SESSION = ()
    SESSION_ACCESS_LEVEL = ()


@serialize(type_check=coerce)
@dataclass
class ActiveServiceConfig(ConfigItem):
    """ A configuration for an Active service.

    :param type: A unique name of service type, with which it is registered into the system.
    :type type: str

    :param name: A unique name of service on a given node.

        This name currently equals to a port identification, but in future releases a service will have a list of opened
        ports.
    :type name: str

    :param owner: An identity of the owner of the service.

        This identity should exist on the node, but it is currently not enforced.
    :type owner: str

    :param access_level: The access level of the service, i.e., the resources the service can access.
    :type access_level: AccessLevel

    :param configuration: A dictionary of any parameters configuring the behavior of the service.

        The configuration is strictly service-dependent and there is no common configuration. For possible values
        consult service documentation.
    :type configuration: Optional[Dict[str, Any]]

    """
    type: str
    name: str
    owner: str
    access_level: AccessLevel
    configuration: Optional[Dict[str, Any]] = None
    id: str = ""
    ref: str = field(default_factory=lambda: str(uuid4()))


@serialize(type_check=coerce)
@dataclass
class PassiveServiceConfig(ConfigItem):
    """ A configuration for a Passive service.

        Any configuration parameter in the form Union[...Config, str] means that the parameter can be specified either
        inline, or can be referenced by its id. This enables a better structuring of a code.

    :param name: A name of the passive service.

        While the name can be anything, and it is not in any way enforced, this name is used to evaluate applicability
        of exploits, so two services susceptible to the same exploit should have the same name.

        This name currently equals to a port identification, but in future releases a service will have a list of opened
        ports.
    :type name: str

    :param owner: An identity of the owner of the service.

        This identity should exist on the node, but it is currently not enforced.
    :type owner: str

    :param version: A version of the service.

        The version specifier should follow semantic versioning scheme. Otherwise unintended consequences will occur.
        The version is also used to evaluate applicability of exploits.
    :type version: str

    :param local: Availability of the service across the network.

        If TRUE, the service does not have associated network port and can't be accessed remotely.
        If FALSe, the service does have a network port with its name associated and can be accessed, provided the
        routing enables it.
    :type local: bool

    :param access_level: The access level of the service.
        If the service is successfully compromised, then this will be the access level that the attacker will be able to
        use.
    :type access_level: AccessLevel

    :param authentication_providers: A set of authentication providers that the service uses.

        Example:

        .. code-block:: python

            local_password_auth = AuthenticationProviderConfig(
                provider_type=AuthenticationProviderType.LOCAL,
                token_type=AuthenticationTokenType.PASSWORD,
                token_security=AuthenticationTokenSecurity.SEALED,
                timeout=30
            )

            c = PassiveServiceConfig(
                ...
                authentication_providers=[local_password_auth("openssh_local_pwd_auth")],
                ...
            )

    :type authentication_providers: List[Optional[Union[AuthenticationProviderConfig, str]]]

    :param access_schemes: A set of schemes for accessing the service.

        An access scheme consists of authentication providers and authorization domains. For the complete logic behind
        authentication and authorization, see developer documentation.

        Example:

        .. code-block:: python

            access_schemes=[AccessSchemeConfig(
                authentication_providers=["openssh_local_pwd_auth"],
                authorization_domain=AuthorizationDomainConfig(
                    type=AuthorizationDomainType.LOCAL,
                    authorizations=[
                        AuthorizationConfig("user1", AccessLevel.LIMITED, id="ssh_auth_1"),
                        AuthorizationConfig("user2", AccessLevel.LIMITED, id="ssh_auth_2"),
                        AuthorizationConfig("root", AccessLevel.ELEVATED)
                    ]
                )
            )]

    :type access_schemes: List[AccessSchemeConfig]

    :param public_data: A set of data, that can be extracted from the service without authorization.
    :type public_data: List[Union[DataConfig, str]]

    :param private_data: A set of data, that can be extracted from the service with authorization.
    :type private_data: List[Union[DataConfig, str]]

    :param public_authorizations: A set of authentications/authorizations that can be extracted from the service without authorization.

        This function is DEPRECATED, because it does not conform to new authentication/authorization framework, but it
        is currently left here, because some tests still depend on it.
    :type public_authorizations: List[Union[AuthorizationConfig, str]]

    :param public_authorizations: A set of authentications/authorizations that can be extracted from the service with authorization.

        This function is DEPRECATED, because it does not conform to new authentication/authorization framework, but it
        is currently left here, because some tests still depend on it.
    :type public_authorizations: List[Union[AuthorizationConfig, str]]

    :param parameters: A set of parameters for the service.

        The parameters can be chosen from the specific domain of ServiceParameter.

        Example:

        .. code-block:: python

            parameters=[
                (ServiceParameter.ENABLE_SESSION, True),
                (ServiceParameter.SESSION_ACCESS_LEVEL, AccessLevel.LIMITED)
            ],

    :type parameters: List[Tuple[ServiceParameter, Any]]
    """
    name: str
    owner: str
    version: str
    local: bool
    access_level: AccessLevel
    authentication_providers: List[Optional[Union[AuthenticationProviderConfig, str]]] = field(default_factory=lambda: [])
    access_schemes: List[AccessSchemeConfig] = field(default_factory=lambda: [])
    public_data: List[Union[DataConfig, str]] = field(default_factory=lambda: [])
    private_data: List[Union[DataConfig, str]] = field(default_factory=lambda: [])
    public_authorizations: List[Union[AuthorizationConfig, str]] = field(default_factory=lambda: [])
    private_authorizations: List[Union[AuthorizationConfig, str]] = field(default_factory=lambda: [])
    parameters: List[Tuple[ServiceParameter, Any]] = field(default_factory=lambda: [])
    id: str = ""
    ref: str = field(default_factory=lambda: str(uuid4()))



from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import List, Optional, Tuple
from netaddr import IPAddress, IPNetwork


class FirewallPolicy(Enum):
    """
    Represents a firewall policy, which dictates how network traffic is handled.

    Possible values:
        :ALLOW: The traffic is allowed to pass.
        :DENY: The traffic is dropped.
    """
    ALLOW = 0,
    DENY = 1


@dataclass
class FirewallRule:
    """
    Represents a firewall rule, which applies a firewall policy based on source and destination.

    :param src_net: A source network of a traffic.
    :type src_net: IPNetwork

    :param dst_net: A destination network of a traffic.
    :type dst_net: IPNetwork

    :param service: A name of the destination service.
    :type service: str

    :param policy: A policy that should be applied.
    :type policy: FirewallPolicy
    """
    src_net: IPNetwork
    dst_net: IPNetwork
    service: str
    policy: FirewallPolicy


class FirewallChainType(Enum):
    """
    A firewall chain represents a set of rules that are applied to a network traffic. The type depends on the direction
    of the traffic.

    Possible values:
        :INPUT: The traffic originates at the same node as the firewall.
        :OUTPUT: The traffic is destined to the same node as the firewall.
        :FORWARD: The traffic is passing through the same node as the firewall.
    """
    INPUT = 0,
    OUTPUT = 1,
    FORWARD = 2


class Firewall(ABC):
    """
    Firewall is represented as a collection of chains, with a default policy that is applied, unless specified
    otherwise.
    """
    @abstractmethod
    def list_rules(self, chain: Optional[FirewallChainType] = None) -> List[Tuple[FirewallChainType, FirewallPolicy,
                                                                                  List[FirewallRule]]]:
        """
        List rules registered in the firewall.

        :param chain: If specified, lists only rules associated with the given chain.
        :type chain: Optional[FirewallChainType]

        :return: A collection of rules with associated chain and the default policy.
        """

    @abstractmethod
    def add_local_ip(self, ip: IPAddress) -> None:
        """
        Add an IP address that is considered local for the purpose of determining the chain of processed traffic.

        :param ip: The IP address to add.
        :type ip: IPAddress
        """

    @abstractmethod
    def remove_local_ip(self, ip: IPAddress) -> None:
        """
        Remove an IP address that is considered local for the purpose of determining the chain of processed traffic.

        :param ip: The IP address to remove.
        :type ip: IPAddress
        """

    @abstractmethod
    def add_rule(self, chain: FirewallChainType, rule: FirewallRule) -> None:
        """
        Adds a rule to a given chain.

        :param chain: The chain to apply the rule to.
        :type chain: FirewallChainType

        :param rule: The rule to add.
        :type rule: FirewallRule
        """

    @abstractmethod
    def remove_rule(self, chain: FirewallChainType, index: int) -> None:
        """
        Removes a rule from a given chain.

        :param chain: The chain to remove the rule from.
        :type chain: FirewallChainType

        :param index: An index of the rule within the chain.
        :type index: int
        """

    @abstractmethod
    def set_default_policy(self, chain: FirewallChainType, policy: FirewallPolicy) -> None:
        """
        Sets a default policy for a given chain.

        :param chain: The chain to set default policy at.
        :type chain: FirewallChainType

        :param policy: The default policy to set.
        :type policy: FirewallPolicy
        """

    @abstractmethod
    def get_default_policy(self, chain: FirewallChainType) -> FirewallPolicy:
        """
        Gets the default policy for a given chain.

        :param chain: The chain to get the policy from.
        :type chain: FirewallChainType

        :return: The default policy.
        """

    @abstractmethod
    def evaluate(self, src_ip: IPAddress, dst_ip: IPAddress, dst_service: str) -> Tuple[bool, int]:
        """
        Evaluates, whether a traffic with given source and destination will pass through or not, according to the rules.

        :param src_ip: The source IP of the traffic.
        :type src_ip: IPAddress

        :param dst_ip: The destination IP of the traffic.
        :type dst_ip: IPAddress

        :param dst_service: The destination service of the traffic.
        :type dst_service: str

        :return: A tuple indicating if the traffic should pass and how long the processing took.
        """



import logging

from typing import Any, Dict, Optional, Tuple

from netaddr import IPAddress

from cyst.api.environment.message import Message, MessageType, Request, Response, Status, StatusOrigin, StatusValue
from cyst.api.environment.messaging import EnvironmentMessaging
from cyst.api.environment.resources import EnvironmentResources
from cyst.api.host.service import ActiveService, ActiveServiceDescription


class ReverseShell(ActiveService):

    def __init__(self,
                 msg: EnvironmentMessaging,
                 res: EnvironmentResources,
                 id: str,
                 args: Optional[Dict[str, Any]] = None) -> None:
        self._messaging = msg
        self._resources = res
        self._id = id
        self._log = logging.getLogger("services.reverse_shell")

        self._success_sent = False
        if args and (origin := args.get("origin")):
            self._origin_request: Request = origin
        else:
            raise ValueError("Reverse shell requires the request which inicialized it")

        self._delay: int = delay if args and (delay := args.get("delay")) else 0
        self._ignore_requests = args is None or args.get("ignore_requests", True)

        if args and (target := args.get("target")):
            self._target: Tuple[IPAddress, str] = target  # (ip, service)
        else:
            raise ValueError("Reverse shell requires target parameter")

        if not (action := res.action_store.get("cyst:active_service:open_session")):
            raise KeyError("Action 'open_session' is not present in action store")
        self._open_session = action

    async def run(self) -> None:
        self._log.info("Launched a reverse shell service")
        self._log.debug(f"Sending intial request to {self._target}")
        self._request_open_session(delay=0)

    async def process_message(self, message: Message) -> Tuple[bool, int]:
        self._log.debug(f"Processing message {message.id} : {message}")

        is_from_target = (message.src_ip, message.src_service) == self._target

        if message.type is MessageType.RESPONSE and is_from_target:
            response = message.cast_to(Response)

            if not self._success_sent:
                if response.status.value is StatusValue.SUCCESS:
                    self._log.debug(f"Sending final response back to {self._origin_request.id}")
                    self._send_success_to_origin()
                else:
                    self._respond_with_error(self._origin_request, "Failed to open session")

            self._log.debug(f"Sending another open session request to {self._target}")
            self._request_open_session(self._delay)
            return True, 1

        if message.type is MessageType.REQUEST and not self._ignore_requests:
            self._respond_with_error(message.cast_to(Request), "Service doesn't accept requests")

        return False, 1

    def _request_open_session(self, delay: int) -> None:
        dst_address, dst_service = self._target
        request = self._messaging.create_request(dst_address, dst_service, self._open_session)
        self._messaging.send_message(request, delay)

    def _respond_with_error(self, request: Request, error: str) -> None:
        response = self._messaging.create_response(
            request, Status(StatusOrigin.SERVICE, StatusValue.FAILURE), error)
        self._messaging.send_message(response)

    def _send_success_to_origin(self) -> None:
        request = self._origin_request
        response = self._messaging.create_response(request,
                                                   Status(StatusOrigin.SERVICE,
                                                          StatusValue.SUCCESS),
                                                   f"Session opened as a reaction to {self._origin_request.id}",
                                                   session=request.session)
        self._messaging.send_message(response)
        self._success_sent = True


def create_shell(msg: EnvironmentMessaging, res: EnvironmentResources,
                 id:str, args: Optional[Dict[str, Any]]) -> ActiveService:
    return ReverseShell(msg, res, id, args)


service_description = ActiveServiceDescription(
    "reverse_shell",
    "A service acting as a reverse shell. It will send an 'open_session' action to the given endpoint",
    create_shell)



from netaddr import IPNetwork, IPAddress

from cyst.api.environment.environment import Environment
from cyst.api.configuration import NodeConfig, PassiveServiceConfig, AccessLevel, ExploitConfig, VulnerableServiceConfig,\
                                   RouterConfig, InterfaceConfig, ConnectionConfig, ActiveServiceConfig
from cyst.api.host.service import Service
from cyst.api.logic.exploit import ExploitLocality, ExploitCategory

from cyst_services.scripted_actor.main import ScriptedActorControl

target = NodeConfig(
    active_services=[],
    passive_services=[
        PassiveServiceConfig(
            type="bash",
            owner="root",
            version="8.1.0",
            access_level=AccessLevel.LIMITED,
            local=True,
            id="bash_service"
        ),
        PassiveServiceConfig(
            type="lighttpd",
            owner="www",
            version="1.4.62",
            access_level=AccessLevel.LIMITED,
            local=False,
            id="web_server"
        )
    ],
    shell="bash",
    interfaces=[],
    traffic_processors=[],
    id="target"
)

# ----------------------------------------------------------------------------------------------------------------------
# Addition to 04_create_network.py
# ----------------------------------------------------------------------------------------------------------------------
attacker = NodeConfig(
    active_services=[
        ActiveServiceConfig(
            type="scripted_actor",
            name="attacker",
            owner="attacker",
            access_level=AccessLevel.LIMITED,
            id="attacker_service"
        )
    ],
    passive_services=[],
    interfaces=[],
    traffic_processors=[],
    shell="",
    id="attacker_node"
)
# ----------------------------------------------------------------------------------------------------------------------

exploit1 = ExploitConfig(
    services=[
        VulnerableServiceConfig(
            name="lighttpd",
            min_version="1.4.62",
            max_version="1.4.62"
        )
    ],
    locality=ExploitLocality.REMOTE,
    category=ExploitCategory.CODE_EXECUTION,
    id="http_exploit"
)

router = RouterConfig(
    interfaces=[
      InterfaceConfig(
        ip=IPAddress("192.168.0.1"),
        net=IPNetwork("192.168.0.1/24"),
        index=0
      ),
      InterfaceConfig(
        ip=IPAddress("192.168.0.1"),
        net=IPNetwork("192.168.0.1/24"),
        index=1
      )
    ],
    traffic_processors=[],
    id="router"
)

connection1 = ConnectionConfig(
        src_id="target",
        src_port=-1,
        dst_id="router",
        dst_port=0
)

# ----------------------------------------------------------------------------------------------------------------------
# Addition to 04_create_network.py
# ----------------------------------------------------------------------------------------------------------------------
connection2 = ConnectionConfig(
        src_id="attacker_node",
        src_port=-1,
        dst_id="router",
        dst_port=1
)

e = Environment.create().configure(target, exploit1, router, connection1, attacker, connection2)

attacker_service = e.configuration.general.get_object_by_id("attacker_node.attacker", Service).active_service
attacker_control = e.configuration.service.get_service_interface(attacker_service, ScriptedActorControl)
# ----------------------------------------------------------------------------------------------------------------------

e.control.init()
e.control.run()
e.control.commit()

stats = e.infrastructure.statistics
print(f"Run id: {stats.run_id}\nStart time real: {stats.start_time_real}\n"
      f"End time real: {stats.end_time_real}\nDuration virtual: {stats.end_time_virtual}")



import logging
from typing import Callable

cyst_defaults = [
    "virtual_time",
    "run_id"
]


class CYSTFormatter(logging.Formatter):
    def format(self, record):
        for k, v in self._cyst_defaults.items():
            if isinstance(v, Callable):
                record.__setattr__(k, v())
            else:
                record.__setattr__(k, v)
        return super().format(record)

    def __init__(self, fmt=None, datefmt=None, style="%", validate=True, *, defaults=None):
        self._cyst_defaults = {}

        for k in cyst_defaults:
            if k in defaults:
                self._cyst_defaults[k] = defaults.get(k)

        super().__init__(fmt, datefmt, style, validate, defaults=defaults)




from dataclasses import dataclass, field
from typing import List, Union
from uuid import uuid4
from serde import serialize, coerce

from cyst.api.configuration.configuration import ConfigItem
from cyst.api.configuration.network.elements import ConnectionConfig
from cyst.api.configuration.network.node import NodeConfig
from cyst.api.configuration.network.router import RouterConfig

# TODO: Useless - remove
@serialize(type_check=coerce)
@dataclass
class NetworkConfig(ConfigItem):
    """ Configuration of a network.

    A network is a collection of nodes linked by connections. The configuration enables creation of unconnected subnets,
    but to make any use of it, there needs to be added a set of actions, which enable setting up/tearing down of
    connections.

    :param nodes: A collection of nodes and routers that constitute the vertices of network topology.
    :type nodes: List[Union[NodeConfig, RouterConfig, str]]

    :param connections: A collection of connections that constitute the edges of network topology.
    :type connections: List[Union[ConnectionConfig, str]]
    """
    nodes: List[Union[NodeConfig, RouterConfig, str]]
    connections: List[Union[ConnectionConfig, str]]
    ref: str = field(default_factory=lambda: str(uuid4()))
    name: str = "__network"
    id: str = ""



from __future__ import annotations

from typing import List, Any

from cyst.api.environment.configuration import ActionConfiguration
from cyst.api.logic.action import ActionParameterDomain

from cyst.core.logic.action import ActionParameterDomainImpl


class ActionConfigurationImpl(ActionConfiguration):

    def create_action_parameter_domain_any(self) -> ActionParameterDomain:
        return ActionParameterDomainImpl()

    def create_action_parameter_domain_range(self, default: int, min: int, max: int,
                                             step: int = 1) -> ActionParameterDomain:
        return ActionParameterDomainImpl.bind_range(default, min, max, step)

    def create_action_parameter_domain_options(self, default: Any, options: List[Any]) -> ActionParameterDomain:
        return ActionParameterDomainImpl.bind_options(default, options)



import asyncio
from typing import List, Optional, Tuple, Any, Dict

from cyst.api.logic.action import Action, ActionDescription, ActionParameter,ActionParameterDomain, ActionParameterDomainType, ActionType
from cyst.api.logic.exploit import Exploit


class ActionParameterDomainImpl(ActionParameterDomain):

    def __init__(self, type: ActionParameterDomainType = ActionParameterDomainType.ANY,
                 range_min: int = -1, range_max: int = -1, range_step: int = -1, options: List[Any] = None,
                 default: Any = None):
        self._type = type
        self._default = default
        self._range_min = range_min
        self._range_max = range_max
        self._range_step = range_step
        if not options:
            self._options = []
        else:
            self._options = options

    @classmethod
    def bind_range(cls, default: int, range_min: int, range_max: int, range_step: int = 1) -> 'ActionParameterDomainImpl':
        return cls(ActionParameterDomainType.RANGE, range_min=range_min, range_max=range_max, range_step=range_step, default=default)

    @classmethod
    def bind_options(cls, default: Any, options: List[Any]) -> 'ActionParameterDomainImpl':
        return cls(ActionParameterDomainType.OPTIONS, options=options, default=default)

    @property
    def type(self) -> ActionParameterDomainType:
        return self._type

    @property
    def range_min(self) -> int:
        if self._type != ActionParameterDomainType.RANGE:
            raise AttributeError("Attempting to get lower range bound on a non-range domain")
        return self._range_min

    @property
    def range_max(self) -> int:
        if self._type != ActionParameterDomainType.RANGE:
            raise AttributeError("Attempting to get upper range bound on a non-range domain")
        return self._range_max

    @property
    def range_step(self) -> int:
        if self._type != ActionParameterDomainType.RANGE:
            raise AttributeError("Attempting to get range step on a non-range domain")
        return self._range_min

    @property
    def options(self) -> List[Any]:
        if self._type != ActionParameterDomainType.OPTIONS:
            raise AttributeError("Attempting to get options on a non-option domain")
        return self._options

    def validate(self, value: Any) -> bool:
        if self._type == ActionParameterDomainType.ANY:
            return True

        if self._type == ActionParameterDomainType.RANGE:
            if not isinstance(value, int):
                return False

            if value < self._range_min or value > self._range_max or (value - self._range_min) % self._range_step != 0:
                return False

            return True

        if self._type == ActionParameterDomainType.OPTIONS:
            if value in self._options:
                return True

            return False

        return NotImplemented #MYPY Complains
    def default(self) -> Any:
        return self._default

    def __getitem__(self, item: int) -> Any:
        if self._type == ActionParameterDomainType.ANY:
            raise IndexError("Attempting to get value from unbounded domain")

        # __getitem__ gets item-th element from the number range
        if self._type == ActionParameterDomainType.RANGE:
            return self._range_min + item * self._range_step

        if self._type == ActionParameterDomainType.OPTIONS:
            return self._options[item]

    def __len__(self) -> int:
        if self._type == ActionParameterDomainType.ANY:
            raise ValueError("Unbounded domain has no length")

        if self._type == ActionParameterDomainType.RANGE:
            return (self._range_max - self._range_min) // self._range_step

        if self._type == ActionParameterDomainType.OPTIONS:
            return len(self._options)

        return NotImplemented #MYPY Complains


class ActionImpl(Action):

    def __init__(self, action: ActionDescription):
        self._type = action.type
        self._id = action.id
        fragments = action.id.split(":")
        self._namespace = fragments[0]
        self._fragments = fragments[1:]
        self._description = action.description
        self._exploit: Optional[Exploit] = None
        self._parameters: Dict[str, ActionParameter] = {}
        for p in action.parameters:
            self._parameters[p.name] = p
        self._components: List[Action] = []

    def __getstate__(self) -> dict:
        return {
            "_type": self._type,
            "_id": self._id,
            "_description": self._description,
            "_exploit": self._exploit,
            "_parameters": self._parameters,
            "_components": self._components
        }

    def __setstate__(self, state) -> None:
        self._type = state["_type"]
        self._id = state["_id"]
        fragments = self._id.split(":")
        self._namespace = fragments[0]
        self._fragments = fragments[1:]
        self._description = state["_description"]
        self._exploit = state["_exploit"]
        self._parameters = state["_parameters"]
        self._components = state["_components"]

    @property
    def id(self) -> str:
        return self._id

    @property
    def type(self) -> ActionType:
        return self._type

    @property
    def description(self) -> str:
        return self._description

    @property
    def namespace(self) -> str:
        return self._namespace

    @property
    def fragments(self) -> List[str]:
        return self._fragments

    @property
    def exploit(self) -> Optional[Exploit]:
        return self._exploit

    def set_exploit(self, exploit: Optional[Exploit]) -> None:
        self._exploit = exploit

    @property
    def parameters(self) -> Dict[str, ActionParameter]:
        return self._parameters

    def add_parameters(self, *params: ActionParameter) -> None:
        for p in params:
            self._parameters[p.name] = p

    @property
    def components(self) -> List[Action]:
        # TODO: This enables modification of components in a non-standard way. Keeping this comment here, because
        #   this is a problem that appears elsewhere.
        return self._components

    def add_components(self, *components: Action) -> None:
        for c in components:
            if isinstance(c, ActionImpl):
                if c._type != ActionType.COMPONENT:
                    raise RuntimeError(f"Cannot add non-component action as an action component. Provided type for action {c.id} is {c._type}")
                else:
                    self._components.append(c)

    @staticmethod
    def cast_from(o: Action) -> 'ActionImpl':
        if isinstance(o, ActionImpl):
            return o
        else:
            raise ValueError("Malformed underlying object passed with the Action interface")

    def copy(self):
        return ActionImpl(ActionDescription(self.id, self.type, self._description, list(self._parameters.values())))


class CompositeAction:
    def __init__(self):
        self._future = None

    def id(self):
        return 1

    def __await__(self):
        return self.execute().__await__()

    async def call_action(self, action_id: str):
        pass

    async def execute(self):
        print("Doing some action")
        await self._future
        print("Action done")


class ScanTheNet(CompositeAction):
    async def execute(self):
        result_set = []
        for i in range(3):
            result_set.append(await self.call_action("scan:the:machine"))
        return result_set


class CompositeActionManager:
    def __init__(self):
        self._loop = asyncio.new_event_loop()

    def process_request(self, composite_action: CompositeAction):
        future = self._loop.create_future()
        composite_action.set_loop(self._loop)
        composite_action.execute(future)


    def process_response(self):
        pass


from __future__ import annotations

import asyncio
from datetime import datetime
from heapq import heappush
from time import struct_time, localtime
from typing import TYPE_CHECKING, Any, Optional

from cyst.api.environment.external import ExternalResources
from cyst.api.environment.clock import Clock
from cyst.api.environment.resources import EnvironmentResources
from cyst.api.environment.platform import PlatformSpecification
from cyst.api.environment.stats import Statistics
from cyst.api.environment.stores import ExploitStore, ActionStore
from cyst.api.host.service import ActiveService
from cyst.api.utils.counter import Counter

from cyst.core.environment.stats import StatisticsImpl
from cyst.core.environment.stores import ActionStoreImpl, ServiceStoreImpl, ExploitStoreImpl
from cyst.core.environment.external_resources import ExternalResourcesImpl

if TYPE_CHECKING:
    from cyst.core.environment.environment import _Environment


# ----------------------------------------------------------------------------------------------------------------------
class EnvironmentResourcesImpl(EnvironmentResources):
    def __init__(self, env: _Environment, platform: Optional[PlatformSpecification] = None):
        self._env = env
        self._action_store = ActionStoreImpl(platform)
        self._exploit_store = ExploitStoreImpl()
        self._clock = None
        self._external_resources = None

    def init_resources(self, loop: asyncio.AbstractEventLoop, clock: Clock):
        self._clock = clock
        self._external_resources = ExternalResourcesImpl(loop, clock)

    @property
    def action_store(self) -> ActionStore:
        return self._action_store

    @property
    def exploit_store(self) -> ExploitStore:
        return self._exploit_store

    @property
    def clock(self) -> Clock:
        return self._clock

    @property
    def external(self) -> ExternalResources:
        return self._external_resources



from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum, auto
from dataclasses import dataclass
from typing import List, NamedTuple, Optional
from flags import Flags
from netaddr import IPAddress


@dataclass
class Event:
    """
    Event represents an identification of a phenomenon within a message. Typically it can be some alert from an IDS,
    which the mechanisms downstream can act upon.

    :param id: An identifier of the event.
    :type id: str
    """
    id: str


class TCPFlags(Flags):
    """
    TCP protocol flags. Can be used separately or combined together by means of the OR operator. Example:

    .. code-block:: python

        f1 = TCPFlags.R
        f2 = TCPFlags.S | TCPFlags.A

    Possible values:
        :S: SYN
        :A: ACK
        :R: RESET
        :P: PUSH
        :U: UPDATE
        :F: FIN

    """
    S = ()
    A = ()
    R = ()
    P = ()
    U = ()
    F = ()


class Protocol(Enum):
    """
    A communication protocol used in the messages.

    Possible values:
        :UDP: UDP Protocol.
        :TCP: TCP Protocol.
        :ICMP: ICMP Protocol.
    """
    UDP = auto()
    TCP = auto()
    ICMP = auto()


class FlowDirection(Enum):
    """
    Direction of the flow.

    Possible values:
        :REQUEST: The flow was create from a message of the request type.
        :RESPONSE: The flow was created from a message of the response type.
    """
    REQUEST = auto()
    RESPONSE = auto()


@dataclass
class Flow:
    """
    Flow (in network terminology) represents an aggregation of packets. However, because the simulation engine is not
    modelling the communication at the packet level, but rather a flow-level, the flow describes one message. It adds
    statistical and L2/L3 layer information on top of abstract message description.

    :param id: Identifier of the flow. It is the same as the ID of the message it was created from. TODO: There is a
        discrepancy, as the message id is int. What the hell is the purpose of Flow id?
    :type id: str

    :param direction: A direction of the flow.
    :type direction: FlowDirection

    :param packet_count: A number of packets aggregated into the flow.
    :type packet_count: int

    :param duration: A number of simulated time units it took for the flow to realize.
    :type duration: int

    :param flags: A set of TCP flags.
    :type flags: TCPFlags

    :param protocol: A protocol type.
    :type protocol: Protocol
    """
    id: str
    direction: FlowDirection
    packet_count: int
    duration: int
    flags: TCPFlags
    protocol: Protocol


@dataclass
class Metadata:
    """
    Metadata represents information that could be observed if one took a look at the packets that were exchanged
    between message source and destination. Metadata are supplied to messages by metadata providers. For more info
    see :class:`cyst.api.environment.metadata_provider.MetadataProvider`. Because of the reliance on providers,
    not all metadata need to be specified.

    :param src_ip: A source IP of the message. Does not necessarily have to be the same as message source IP, as the
        provider can put there anything it thinks is the source (e.g., seeing through the proxy or NAT).
    :type src_ip: Optional[IPAddress]

    :param dst_ip: A destination IP of the message. Does not necessarily have to be the same as message destination IP,
        as the provider can put there anything it thinks is the ultimate destination.
    :type dst_ip: Optional[IPAddress]

    :param dst_service: A destination service of the message. Does not necessarily have to be the same as message
        destination service, as the provider can put there anything it thinks is the ultimate destination (e.g.,
        final result of port knocking, understanding port forwarding).
    :type dst_service: Optional[str]

    :param event: A recognized event.
    :type event: Optional[str]

    :param flows: A list of flows associated with the message.
    :type flows: Optional[List[Flow]]
    """
    src_ip: Optional[IPAddress] = None
    dst_ip: Optional[IPAddress] = None
    dst_service: Optional[str] = None
    event: Optional[str] = None
    flows: Optional[List[Flow]] = None



from __future__ import annotations

import asyncio
import uuid
import time

from contextlib import suppress
from typing import Tuple, TYPE_CHECKING

from cyst.api.environment.control import EnvironmentState, EnvironmentControl
from cyst.api.environment.message import ComponentState
from cyst.api.environment.platform_specification import PlatformType
from cyst.api.environment.stats import Statistics
from cyst.api.host.service import ServiceState

from cyst.core.environment.serialization import Serializer
from cyst.core.environment.stats import StatisticsImpl

if TYPE_CHECKING:
    from cyst.core.environment.environment import _Environment


class EnvironmentControlImpl(EnvironmentControl):
    def __init__(self, env: _Environment):
        self._env = env

    @property
    def state(self) -> EnvironmentState:
        return _state(self._env)

    def init(self) -> Tuple[bool, EnvironmentState]:
        return _init(self._env)

    def commit(self) -> None:
        return _commit(self._env)

    def reset(self, run_id: str = str(uuid.uuid4())) -> Tuple[bool, EnvironmentState]:
        return _reset(self._env, run_id)

    def run(self) -> Tuple[bool, EnvironmentState]:
        return _run(self._env)

    def pause(self) -> Tuple[bool, EnvironmentState]:
        return _pause(self._env)

    def terminate(self) -> Tuple[bool, EnvironmentState]:
        return _terminate(self._env)

    def add_pause_on_request(self, id: str) -> None:
        return _add_pause_on_request(self._env, id)

    def remove_pause_on_request(self, id: str) -> None:
        return _remove_pause_on_request(self._env, id)

    def add_pause_on_response(self, id: str) -> None:
        return _add_pause_on_response(self._env, id)

    def remove_pause_on_response(self, id: str) -> None:
        return _remove_pause_on_response(self._env, id)

    def snapshot_save(self) -> str:
        return _snapshot_save(self._env)

    def snapshot_load(self, state: str) -> None:
        return _snapshot_load(self._env, state)

    def transaction_start(self) -> Tuple[int, int, str]:
        return _transaction_start(self._env)

    def transaction_commit(self, transaction_id: int) -> Tuple[bool, str]:
        return _transaction_commit(self._env, transaction_id)

    def transaction_rollback(self, transaction_id: int) -> Tuple[bool, str]:
        return _transaction_rollback(self._env, transaction_id)


# ----------------------------------------------------------------------------------------------------------------------
# Free function implementations of the above class. It is being done this way to shut up the type checking and to
# overcome python's limitation on having a class implemented in multiple files.
def _state(self: _Environment) -> EnvironmentState:
    return self._state


def _init(self: _Environment) -> Tuple[bool, EnvironmentState]:
    if self._initialized:
        return True, self._state

    if self._state == EnvironmentState.RUNNING or self._state == EnvironmentState.PAUSED:
        return False, self._state

    self._pause = False
    self._terminate = False
    self._state = EnvironmentState.INIT

    # Set basic statistics
    s = StatisticsImpl.cast_from(self.infrastructure.statistics)
    s.configuration_id = self._runtime_configuration.config_id
    s.start_time_real = time.time()

    # Initialize the platform, if needed
    if self._platform:
        self._platform.init()

    self._initialized = True

    return True, self._state


def _reset(self: _Environment, run_id: str = str(uuid.uuid4())) -> Tuple[bool, EnvironmentState]:
    if self._state != EnvironmentState.FINISHED and self._state != EnvironmentState.TERMINATED:
        return False, self._state

    self._network.reset()
    self._time = 0
    self._start_time = time.localtime()
    self._message_queue.clear()
    self._pause = False
    self._terminate = False
    self._run_id = run_id
    self._state = EnvironmentState.INIT

    return True, self._state

def _run(self: _Environment) -> Tuple[bool, EnvironmentState]:

    if not self._initialized:
        return False, self._state

    # if paused, unpause
    if self._state == EnvironmentState.PAUSED:
        self._pause = False

    # Run
    self._finish = False

    service_run_tasks = set()
    # As a first thing after init, we run all the services (and make sure we don't run it multiple times)
    if self._state == EnvironmentState.INIT and len(service_run_tasks) == 0:
        for service in self._service_store.get_active_services():
            t = self._loop.create_task(service.run())
            service_run_tasks.add(t)
            t.add_done_callback(service_run_tasks.discard)

    while not (self._pause or self._finish or self._terminate):
        # Make sure to not terminate the pump before the run() methods finish
        if self._state == EnvironmentState.INIT:
            if len(service_run_tasks) == 0:
                self._state = EnvironmentState.RUNNING

        # And now we are finally doing the ordinary stuff
        t = self._loop.create_task(self._process_async())
        self._loop.call_soon(self._loop.stop)
        self._loop.run_forever()

    if self._pause:
        self._state = EnvironmentState.PAUSED
    else:
        # Send a notification to all active services that we had to terminate
        effect_description = ""
        # Terminate description has a priority
        if self._terminate and self._terminate_reason:
            effect_description = self._terminate_reason
        if not effect_description and self._finish and self._finish_reason:
            effect_description = self._finish_reason

        end_signal = self.messaging.create_signal(signal_origin="__environment",
                                                  state=ComponentState.TERMINATED if self._terminate else ComponentState.FINISHED,
                                                  effect_origin="__environment",
                                                  effect_description=effect_description)

        for service in self._service_store.get_active_services():
            self._loop.create_task(service.process_message(end_signal))

        self._data_store.add_signal(end_signal)

        # Realistically, we need some more loop steps to serve the awaits when all futures are set
        # The number 4 was chosen because it works and gives also a hefty margin. As far as I checked, 1 is enough.
        for _ in range(4):
            self._loop.create_task(self._process_async())
            self._loop.call_soon(self._loop.stop)
            self._loop.run_forever()

        # Terminate clears the task queue and sets the clock back to zero
        if self._terminate:
            self._state = EnvironmentState.TERMINATED
            self._time = 0
            self._message_queue.clear()

        else:
            self._state = EnvironmentState.FINISHED

        # Clear the loop if there are any outstanding tasks
        pending = asyncio.all_tasks(self._loop)
        for task in pending:
            task.cancel()

            with suppress(asyncio.CancelledError):
                self._loop.run_until_complete(task)

    return True, self._state


def _pause(self: _Environment) -> Tuple[bool, EnvironmentState]:

    if self._state != EnvironmentState.RUNNING:
        return False, self._state

    self._pause = True
    # This will return True + running state, but it will be returned to an actor other than the one who called
    # Environment.run() in the first place. Or I hope so...
    return True, self._state


def _terminate(self: _Environment) -> Tuple[bool, EnvironmentState]:

    self._terminate = True
    self._terminate_reason = "Direct call to environment terminate method."

    if self._platform:
        self._platform.terminate()

    return True, self._state


def _commit(self: _Environment) -> None:
    s = StatisticsImpl.cast_from(self.infrastructure.statistics)
    s.end_time_real = time.time()
    s.end_time_virtual = self._time

    self._data_store.add_statistics(self.infrastructure.statistics)

    # TODO: What?
    if self._platform:
        self._platform.terminate()

    if self._runtime_configuration.data_batch_storage and self._runtime_configuration.data_backend != "memory":
        memory = self._data_store.memory

        self._data_store_batch.add_signal(*memory["signals"])
        self._data_store_batch.add_statistics(memory["statistics"])
        self._data_store_batch.add_action(*memory["actions"])
        self._data_store_batch.add_message(*memory["messages"])


def _add_pause_on_request(self: _Environment, id: str) -> None:
    self._pause_on_request.append(id)


def _remove_pause_on_request(self: _Environment, id: str) -> None:
    self._pause_on_request = [x for x in self._pause_on_request if x != id]


def _add_pause_on_response(self: _Environment, id: str) -> None:
    self._pause_on_response.append(id)


def _remove_pause_on_response(self: _Environment, id: str) -> None:
    self._pause_on_response = [x for x in self._pause_on_response if x != id]


def _snapshot_save(self: _Environment) -> str:
    return Serializer.serialize(self)


def _snapshot_load(self: _Environment, state: str) -> None:
    self._replace(Serializer.deserialize(state))


def _transaction_start(self: _Environment) -> Tuple[int, int, str]:
    return (0, 0, "")


def _transaction_commit(self: _Environment, transaction_id: int) -> Tuple[bool, str]:
    return (True, "")


def _transaction_rollback(self: _Environment, transaction_id: int) -> Tuple[bool, str]:
    return (True, "")



import datetime
from typing import Self


class Duration:
    """
    Duration class provides a subjectively more esthetically pleasing thin syntactical wrapper over Python's timedelta.
    Duration is not meant to be instantiated directly from users' code, but through the free functions, which get
    chained with member ones.

    Example:
        > print(secs(2).msecs(37))
        > 0:00:02.037000

    The code that requires the float representation of the whole duration (such as for the sleep function) can use the
    :func:`to_float` method.

    Example:
        time.sleep(secs(2).msecs(37).to_float())

    The available duration specifications mimics those of timedelta:
        :weeks
        :days
        :hours
        :mins [minutes]
        :secs [seconds]
        :msecs [milliseconds]
        :usecs [microsecods]

    For fractions of microseconds (such as hectonanoseconds) just use the appropriate fractional value of any available
    duration.
    """

    def __init__(self):
        self._weeks = 0.0
        self._days = 0.0
        self._hours = 0.0
        self._mins = 0.0
        self._secs = 0.0
        self._msecs = 0.0
        self._usecs = 0.0

    def weeks(self, value: float) -> Self:
        self._weeks = value
        return self

    def days(self, value: float) -> Self:
        self._days = value
        return self

    def hours(self, value: float) -> Self:
        self._hours = value
        return self

    def mins(self, value: float) -> Self:
        self._mins = value
        return self

    def secs(self, value: float) -> Self:
        self._secs = value
        return self

    def msecs(self, value: float) -> Self:
        self._msecs = value
        return self

    def usecs(self, value: float) -> Self:
        self._usecs = value
        return self

    def to_float(self) -> float:
        return self.__repr__().total_seconds()

    def __str__(self):
        return str(self.__repr__())

    def __repr__(self):
        return datetime.timedelta(weeks=self._weeks, days=self._days, hours=self._hours, minutes=self._mins,
                                  seconds=self._secs, milliseconds=self._msecs, microseconds=self._usecs)


def weeks(value: float) -> Duration:
    duration = Duration()
    duration.weeks(value)
    return duration


def days(value: float) -> Duration:
    duration = Duration()
    duration.days(value)
    return duration


def hours(value: float) -> Duration:
    duration = Duration()
    duration.hours(value)
    return duration


def mins(value: float) -> Duration:
    duration = Duration()
    duration.mins(value)
    return duration


def secs(value: float) -> Duration:
    duration = Duration()
    duration.secs(value)
    return duration


def msecs(value: float) -> Duration:
    duration = Duration()
    duration.msecs(value)
    return duration


def usecs(value: float) -> Duration:
    duration = Duration()
    duration.usecs(value)
    return duration



from abc import ABC, abstractmethod


class Data(ABC):
    """
    This class represents an arbitrary data.

    Warning:
        This is currently very underdeveloped and will be overhauled in the near future.
    """

    @property
    @abstractmethod
    def id(self) -> str:
        """
        An identifier of the data that is unique within one service, but preferably across simulation.

        :rtype: str
        """

    @property
    @abstractmethod
    def owner(self) -> str:
        """
        The identity of a user, to whom the data belongs.

        :rtype: str
        """

    @property
    @abstractmethod
    def path(self) -> str:
        """
        The path to a data.

        :rtype: str
        """

    @property
    @abstractmethod
    def description(self) -> str:
        """
        A textual description of the data. Currently, it also serves as a way to store contents of the data, even though
        true contents of the data will never be a part of the simulation.

        :rtype: str
        """



from dataclasses import dataclass
from typing import List

from cyst.api.environment.message import Status, Request, Response


# ----------------------------------------------------------------------------------------------------------------------
# Actions
# Action model combines a request and related responses
@dataclass
class ActionParameterModel:
    name: str
    value: str


@dataclass
class ActionModel:
    # Request part
    message_id: int
    action_id: str
    caller_id: str
    src_ip: str
    dst_ip: str
    dst_service: str
    session_in: str
    auth_in: str
    parameters: List[ActionParameterModel]
    # Response part
    session_out: str = ""
    auth_out: str = ""
    status_origin: str = ""
    status_value: str = ""
    status_detail: str = ""
    response: str = ""

    @classmethod
    def from_request(cls, run_id: str, request: Request):
        parameters = []
        for param in request.action.parameters.values():
            parameters.append(ActionParameterModel(
                name=param.name,
                value=str(param.value)
            ))

        return cls(
            message_id=request.id,
            action_id=request.action.id,
            caller_id=request.platform_specific["caller_id"],
            src_ip=str(request.src_ip) if request.src_ip else "",
            dst_ip=str(request.dst_ip),
            dst_service=request.dst_service,
            session_in=str(request.session.id) if request.session else "",
            auth_in="", # TODO
            parameters=parameters
        )

    def set_response(self, response: Response):
        # This can easily happen, as the request gets assigned source IP in the platform code, after the ActionModel
        # was initialized.
        if not self.src_ip:
            self.src_ip = str(response.dst_ip)
        self.session_out = str(response.session.id) if response.session else ""
        self.auth_out = "" # TODO
        self.status_origin = str(response.status.origin)
        self.status_value = str(response.status.value)
        self.status_detail = str(response.status.detail) if response.status.detail else ""
        self.response = str(response.content)



from __future__ import annotations

import uuid

from typing import TYPE_CHECKING, List, Optional, Any, Type, Dict

from cyst.api.environment.configuration import ServiceConfiguration, ServiceParameter, ActiveServiceInterfaceType
from cyst.api.host.service import PassiveService, Service, ActiveService
from cyst.api.logic.access import AccessScheme, AuthenticationProvider, Authorization, AccessLevel
from cyst.api.logic.data import Data
from cyst.api.network.node import Node
from cyst.api.network.session import Session

from cyst.platform.host.service import PassiveServiceImpl, ServiceImpl
from cyst.platform.logic.data import DataImpl
from cyst.platform.network.node import NodeImpl

if TYPE_CHECKING:
    from cyst.platform.main import CYSTPlatform


class ServiceConfigurationImpl(ServiceConfiguration):
    def __init__(self, platform: CYSTPlatform):
        self._platform = platform

    def create_active_service(self, type: str, owner: str, name: str, node: Node,
                              service_access_level: AccessLevel = AccessLevel.LIMITED,
                              configuration: Optional[Dict[str, Any]] = None, id: str = "") -> Optional[Service]:
        return _create_active_service(self._platform, type, owner, name, node, service_access_level, configuration, id)

    def get_service_interface(self, service: ActiveService,
                              interface_type: Type[ActiveServiceInterfaceType]) -> ActiveServiceInterfaceType:
        if isinstance(service, interface_type):
            return service
        else:
            raise RuntimeError("Given active service does not provide control interface of given type.")

    def create_passive_service(self, type: str, owner: str, version: str = "0.0.0", local: bool = False,
                               service_access_level: AccessLevel = AccessLevel.LIMITED, id: str = "") -> Service:
        return _create_passive_service(self._platform, type, owner, version, local, service_access_level, id)

    def update_service_version(self, service: PassiveService, version: str = "0.0.0") -> None:
        service = PassiveServiceImpl.cast_from(service)
        service.version = version

    def set_service_parameter(self, service: PassiveService, parameter: ServiceParameter, value: Any) -> None:
        service = PassiveServiceImpl.cast_from(service)
        if parameter == ServiceParameter.ENABLE_SESSION:
            service.set_enable_session(value)
        elif parameter == ServiceParameter.SESSION_ACCESS_LEVEL:
            service.set_session_access_level(value)

    def create_data(self, id: Optional[str], owner: str, path: str, description: str) -> Data:
        return _create_data(self._platform, id, owner, path, description)

    def public_data(self, service: PassiveService) -> List[Data]:
        return PassiveServiceImpl.cast_from(service).public_data

    def private_data(self, service: PassiveService) -> List[Data]:
        return PassiveServiceImpl.cast_from(service).private_data

    def public_authorizations(self, service: PassiveService) -> List[Authorization]:
        return PassiveServiceImpl.cast_from(service).public_authorizations

    def private_authorizations(self, service: PassiveService) -> List[Authorization]:
        return PassiveServiceImpl.cast_from(service).private_authorizations

    def sessions(self, service: PassiveService) -> Dict[str, Session]:
        return PassiveServiceImpl.cast_from(service).sessions

    def provides_auth(self, service: PassiveService, auth_provider: AuthenticationProvider) -> None:
        return PassiveServiceImpl.cast_from(service).add_provider(auth_provider)

    def set_scheme(self, service: PassiveService, scheme: AccessScheme) -> None:
        return PassiveServiceImpl.cast_from(service).add_access_scheme(scheme)


# ------------------------------------------------------------------------------------------------------------------
# ServiceConfiguration (with trampoline)
def _create_active_service(self: CYSTPlatform, type: str, owner: str, name: str, node: Node,
                          service_access_level: AccessLevel = AccessLevel.LIMITED,
                          configuration: Optional[Dict[str, Any]] = None, id: str = "") -> Optional[Service]:
    if not id:
        id = NodeImpl.cast_from(node).id + "." + name

    # HACK: A bit of a hack to enable active services to get access to configuration-produced sessions
    sessions: Dict[str, Session] = dict()

    if not configuration:
        configuration = {}
    configuration["__sessions"] = sessions

    srv = self._infrastructure.service_store.create_active_service(type, owner, name, node, service_access_level, configuration, id)
    s = ServiceImpl(type, srv, name, owner, service_access_level, id)

    s.sessions = sessions

    if s:
        self._general_configuration.add_object(id, s)

    return s


def _create_passive_service(self: CYSTPlatform, type: str, owner: str, version: str = "0.0.0", local: bool = False,
                           service_access_level: AccessLevel = AccessLevel.LIMITED, id: str = "") -> Service:
    if not id:
        id = str(uuid.uuid4())

    p = PassiveServiceImpl(type, owner, version, local, service_access_level, id)
    self._general_configuration.add_object(id, p)
    return p


def _create_data(self: CYSTPlatform, id: Optional[str], owner: str, path: str, description: str) -> Data:
    if not id:
        id = str(uuid.uuid4())
    d = DataImpl(id, owner, path, description)
    self._general_configuration.add_object(id, d)
    return d



import logging.config
import uuid
import collections

from dataclasses import dataclass, field
from typing import List, Union, Optional, Dict, Any, Type, Tuple, Callable

import jsonpickle

from cyst.api.environment.configuration import GeneralConfiguration, ObjectType, ConfigurationObjectType
from cyst.api.environment.messaging import EnvironmentMessaging
from cyst.api.environment.platform import Platform
from cyst.api.host.service import ActiveService, PassiveService, Service
from cyst.api.configuration.configuration import ConfigItem
from cyst.api.configuration.host.service import ActiveServiceConfig, PassiveServiceConfig
from cyst.api.configuration.infrastructure.infrastructure import InfrastructureConfig
from cyst.api.configuration.infrastructure.log import LogConfig, LogSource, log_defaults
from cyst.api.configuration.logic.access import AuthorizationConfig, AuthenticationProviderConfig, AccessSchemeConfig, \
    AuthorizationDomainConfig, FederatedAuthorizationConfig
from cyst.api.configuration.logic.data import DataConfig
from cyst.api.configuration.logic.exploit import VulnerableServiceConfig, ExploitParameterConfig, ExploitConfig
from cyst.api.configuration.network.elements import PortConfig, InterfaceConfig, ConnectionConfig, RouteConfig, SessionConfig
from cyst.api.configuration.network.firewall import FirewallChainConfig, FirewallConfig, FirewallPolicy, FirewallChainType
from cyst.api.configuration.network.network import NetworkConfig
from cyst.api.configuration.network.router import RouterConfig
from cyst.api.configuration.network.node import NodeConfig
from cyst.api.environment.environment import Environment
from cyst.api.host.service import AccessLevel
from cyst.api.network.node import Node

from cyst.platform.network.firewall import FirewallImpl
from cyst.platform.network.elements import ConnectionImpl
from cyst.platform.network.node import NodeImpl
from cyst.platform.host.service import ServiceImpl, PassiveServiceImpl, ServiceState


# ----------------------------------------------------------------------------------------------------------------------
# Converting configuration description to actual objects
# ----------------------------------------------------------------------------------------------------------------------
class Configurator:
    def __init__(self, platform: Platform):
        self._platform = platform
        self._refs: Dict[str, Any] = {}
        self._obj_refs: Dict[str, Any] = {}
        self._connections: List[ConnectionConfig] = []
        self._sessions: List[SessionConfig] = []
        self._nodes: List[NodeConfig] = []
        self._routers: List[RouterConfig] = []
        self._active_services: List[ActiveServiceConfig] = []
        self._passive_services: List[PassiveServiceConfig] = []
        self._firewalls: List[FirewallConfig] = []
        self._interfaces: List[Union[InterfaceConfig, PortConfig]] = []
        self._authorizations: List[AuthorizationConfig] = []
        self._data: List[DataConfig] = []
        self._exploits: List[ExploitConfig] = []
        self._authentication_providers: List[AuthenticationProviderConfig] = []
        self._access_schemes: List[AccessSchemeConfig] = []
        self._authorization_domains: List[AuthorizationDomainConfig] = []

    def __getstate__(self):
        result = self.__dict__
        return result

    def __setstate__(self, state):
        self.__dict__.update(state)

    def reset(self):
        self._refs.clear()
        self._obj_refs.clear()
        self._connections.clear()
        self._sessions.clear()
        self._nodes.clear()
        self._routers.clear()
        self._active_services.clear()
        self._passive_services.clear()
        self._firewalls.clear()
        self._interfaces.clear()
        self._authorizations.clear()
        self._data.clear()
        self._exploits.clear()
        self._authentication_providers.clear()
        self._access_schemes.clear()
        self._authorization_domains.clear()

    def get_object_by_id(self, id: str, object_type: Type[ObjectType]) -> ObjectType:

        o = self._obj_refs[id]
        error = False

        # We have to do a bit of a black magic to overcome Service type shadowing
        if isinstance(o, ServiceImpl):
            if object_type is ActiveService and o.active_service:
                return o.active_service
            elif object_type is PassiveService and o.passive_service:
                return o.passive_service
            elif object_type is Service:
                return o
            else:
                error = True
        elif not isinstance(o, object_type):
            error = True

        if error:
            raise AttributeError(
                "Attempting to cast object with id: {} to an incompatible type: {}. Type is {}".format(id,
                                                                                                       str(object_type),
                                                                                                       type(o)))

        return o

    def add_object(self, id: str, obj: Any) -> None:
        self._obj_refs[id] = obj

    def process_cfg_item(self, item: ConfigItem):
        mapping = {
            ConnectionConfig: self._connections,
            SessionConfig: self._sessions,
            RouterConfig: self._routers,
            NodeConfig: self._nodes,
            InterfaceConfig: self._interfaces,
            ActiveServiceConfig: self._active_services,
            PassiveServiceConfig: self._passive_services,
            AuthorizationConfig: self._authorizations,
            DataConfig: self._data,
            FirewallConfig: self._firewalls,
            ExploitConfig: self._exploits,
            AuthenticationProviderConfig: self._authentication_providers,
            AccessSchemeConfig: self._access_schemes,
            AuthorizationDomainConfig: self._authorization_domains,
            FederatedAuthorizationConfig: self._authorizations
        }

        for key, value in item.__dict__.items():
            if not isinstance(value, list):
                value = [value]
            for v in value:
                if type(v) in mapping:
                    self.process_cfg_item(v)

        if type(item) in mapping:
            mapping[type(item)].append(item)  # For once, the type inspection is helpless here

        self._refs[item.ref] = item

    # ------------------------------------------------------------------------------------------------------------------
    # Gather all configuration items
    def configure(self,
                  *configs: Union[NetworkConfig, ConnectionConfig, RouterConfig, NodeConfig,
                                  InterfaceConfig, ActiveServiceConfig, PassiveServiceConfig,
                                  AuthorizationConfig, DataConfig]) -> None:

        # --------------------------------------------------------------------------------------------------------------
        # Store all the configuration items
        # Each type to its own. I am basically exploding the configuration again. It would be better to go top-down,
        # considering it is already pre-configured, but this has to be done in another merge request.
        for config in configs:
            self.process_cfg_item(config)

        # --------------------------------------------------------------------------------------------------------------
        # Build order:
        # 1) Authorizations, Data and Exploits
        # 2) Passive Services
        # 3) Interfaces
        # 4) Nodes and routers
        # 5) Connections

        # 1) Authorizations, Data and Exploits
        # for auth in self._authorizations:
        #    a = self._env.policy.create_authorization(auth.identity, auth.nodes, auth.services, auth.access_level, auth.id)
        #    self._env.policy.add_authorization(a)
        #    self._obj_refs[auth.id] = a

        # Building authentication and authorization infrastructure
        # Prepare authentication tokens based on the authorizations in access schemes
        for scheme in self._access_schemes:
            scheme_instance = self._platform.configuration.access.create_access_scheme(scheme.id)

            # Authorization domain is passed on as a full class from pre-configuration, not as an ID
            # authorization_domain = self._refs[scheme.authorization_domain]
            authorization_domain = scheme.authorization_domain

            # Go through all providers and if they do not exist instantiate them
            for provider_conf in scheme.authentication_providers:
                if isinstance(provider_conf, str):
                    provider_conf: AuthenticationProviderConfig = self._refs[provider_conf]

                if provider_conf.id not in self._obj_refs:
                    provider = self._platform.configuration.access.create_authentication_provider(
                        provider_conf.provider_type,
                        provider_conf.token_type,
                        provider_conf.token_security,
                        provider_conf.ip,
                        provider_conf.timeout,
                        provider_conf.id
                    )
                else:
                    provider = self._obj_refs[provider_conf.id]

                self._platform.configuration.access.add_provider_to_scheme(provider, scheme_instance)

            for auth in authorization_domain.authorizations:
                auth = self._refs[auth.ref]
                if isinstance(auth, AuthorizationConfig) or isinstance(auth, FederatedAuthorizationConfig):
                    identity = auth.identity

                    # This is not very pretty, but the original version that ran this code for each provider caused
                    # problems with the new approach that correctly tracks IDs of objects. In this case objects with
                    # identical IDs could be created multiple times.
                    for provider_conf in scheme.authentication_providers:
                        self._platform.configuration.access.create_and_register_authentication_token(self._obj_refs[provider_conf.id], identity)

                    # WARN: this creates authorizations with services = ["*"], nodes = ["*"] if not Federated is
                    # created, when the authorization process is
                    # done, these are only used as templates so should not be a problem
                    authorization = self._platform.configuration.access.create_authorization(auth.identity,
                                                                                        auth.access_level, auth.id)\
                        if not isinstance(auth, FederatedAuthorizationConfig) else \
                        self._platform.configuration.access.create_authorization(auth.identity, auth.access_level,
                                                                            auth.id, auth.nodes, auth.services)

                    self._platform.configuration.access.add_authorization_to_scheme(authorization, scheme_instance)
                else:
                    raise RuntimeError("Wrong object type provided instead of (Federated)AuthorizationConfig")

        for data in self._data:
            d = self._platform.configuration.service.create_data(data.id, data.owner, data.path, data.description)

        for exploit in self._exploits:
            params = []
            if exploit.parameters:
                for p in exploit.parameters:
                    param = self._platform.configuration.exploit.create_exploit_parameter(p.type, p.value, p.immutable)
                    params.append(param)

            services = []
            for s in exploit.services:
                service = self._platform.configuration.exploit.create_vulnerable_service(s.service, s.min_version,
                                                                                    s.max_version)
                services.append(service)

            e = self._platform.configuration.exploit.create_exploit(exploit.id, services, exploit.locality, exploit.category,
                                                               *params)

            self._platform.configuration.exploit.add_exploit(e)

        # 2) Passive Services
        passive_service_obj = {}
        for service in self._passive_services:
            s = self._platform.configuration.service.create_passive_service(service.name, service.owner, service.version,
                                                                       service.local, service.access_level, service.id)
            # This was pre-split managed in _run method of the environment, but as PassiveServiceImpl is now
            # platform-specific, there is no clear way to set it there.
            PassiveServiceImpl.cast_from(s.passive_service).set_state(ServiceState.RUNNING)

            for d in service.public_data:
                self._platform.configuration.service.public_data(s.passive_service).append(self._obj_refs[d.id])
            for d in service.private_data:
                self._platform.configuration.service.private_data(s.passive_service).append(self._obj_refs[d.id])
            for a in service.public_authorizations:
                self._platform.configuration.service.public_authorizations(s.passive_service).append(self._obj_refs[a.id])
            for a in service.private_authorizations:
                self._platform.configuration.service.private_authorizations(s.passive_service).append(self._obj_refs[a.id])

            for p in service.parameters:
                self._platform.configuration.service.set_service_parameter(s.passive_service, p[0], p[1])

            for prov in service.authentication_providers:
                self._platform.configuration.service.provides_auth(s,
                                                              self._obj_refs[prov.id if isinstance(prov,
                                                                            AuthenticationProviderConfig) else prov])

            for scheme in service.access_schemes:
                self._platform.configuration.service.set_scheme(s.passive_service, self._obj_refs[
                    scheme.id if isinstance(scheme, AccessSchemeConfig) else scheme])

            passive_service_obj[service.id] = s

        # 3) Interfaces
        for iface in self._interfaces:
            if isinstance(iface, PortConfig):
                self._platform.configuration.node.create_port(iface.ip, str(iface.net.netmask), iface.index, iface.id)
            else:
                # TODO: Missing a setting of a gateway (Really todo?)
                self._platform.configuration.node.create_interface(iface.ip, str(iface.net.netmask), iface.index, iface.id)

        # 4) Nodes
        for node in self._nodes:
            n = self._platform.configuration.node.create_node(node.id)
            for i in node.interfaces:
                obj_i = self._obj_refs[i.id]
                self._platform.configuration.node.add_interface(n, obj_i, obj_i.index)

            for service in node.passive_services:
                self._platform.configuration.node.add_service(n, passive_service_obj[service.id])

            for service_cfg in node.active_services:
                s = self._platform.configuration.service.create_active_service(service_cfg.type, service_cfg.owner,
                                                                               service_cfg.name, n, service_cfg.access_level,
                                                                               service_cfg.configuration,
                                                                               service_cfg.id)
                self._platform.configuration.node.add_service(n, s)
            
            for processor_cfg in node.traffic_processors:
                if isinstance(processor_cfg, ActiveServiceConfig):
                    s = self._platform.configuration.service.create_active_service(processor_cfg.type, processor_cfg.owner,
                                                                                   processor_cfg.name, n, processor_cfg.access_level,
                                                                                   processor_cfg.configuration,
                                                                                   processor_cfg.id)
                else:
                    # TODO: Owner/name/access_level not in the config. It should probably be there. Otherwise we have
                    #       to hardcode.
                    s = self._platform.configuration.service.create_active_service("firewall", "root",
                                                                                   "firewall", n, AccessLevel.ELEVATED,
                                                                                   None,
                                                                                   node.id + ".firewall_0")  # TODO: This is definitely wrong in the new naming scheme

                    # This is not very pretty. But coding around it would require unnecessary expansion of the API.
                    if isinstance(s, FirewallImpl):
                        impl: FirewallImpl = s

                        for chain in processor_cfg.chains:
                            impl.set_default_policy(chain.type, chain.policy)
                            for rule in chain.rules:
                                impl.add_rule(chain.type, rule)

                self._platform.configuration.node.add_traffic_processor(n, s)

            self._platform.configuration.node.set_shell(n, n.services.get(node.shell, None))

            self._platform.configuration.network.add_node(n)

        # TODO, HACK: Any source of messages has to be an active service, so that it is registered in the system
        #             and passed the correct EnvironmentMessaging, i.e., we can't just pass the platform one.
        #             Therefore, in this case we just create a dummy active service and access the underlying
        #             environment's messaging. It is unholy, but it will change when the router is turned into
        #             a correct active service. I don't be rewriting the documentation and network API though,
        #             that will be a major pain in the ass.
        dummy_service = self._platform.configuration.service.create_active_service("scripted_actor", "__dummy", "__dummy",
                                                                                   NodeImpl(str(uuid.uuid4())),
                                                                                   AccessLevel.LIMITED,
                                                                                   None, str(uuid.uuid4()))
        platform_messaging: EnvironmentMessaging = dummy_service.active_service._messaging

        for router in self._routers:
            r = self._platform.configuration.node.create_router(router.id, platform_messaging)
            # TODO: This one complains that it is the List of Interface configs, while, after config processing it is
            #       a list of strings. This should probably be fixed to keep it type-happy.
            for iface_cfg in router.interfaces:
                iface = self._obj_refs[iface_cfg.id]
                self._platform.configuration.node.add_interface(r, iface, iface.index)

            for route in router.routing_table:
                route_obj = self._platform.configuration.node.create_route(route.network, route.port, route.metric)
                self._platform.configuration.node.add_route(r, route_obj)

            # TODO: Code duplication... this is one of the things to fix once routers and nodes are combined together

            # HACK: To have a sensible default, we add a permissive firewall that enables forwarding on the router if
            # there are no traffic processors specified. The implementation is not pretty.
            have_fw = False
            for processor_cfg in router.traffic_processors:
                if isinstance(processor_cfg, FirewallConfig):
                    have_fw = True

            if not have_fw:
                fw_id = str(uuid.uuid4())
                cfg = FirewallConfig(
                                        default_policy=FirewallPolicy.DENY,
                                        chains=[
                                            FirewallChainConfig(
                                                type=FirewallChainType.FORWARD,
                                                policy=FirewallPolicy.ALLOW,
                                                rules=[]
                                            )
                                        ]
                                    )
                self._refs[fw_id] = cfg
                router.traffic_processors.append(cfg)

            for processor_cfg in router.traffic_processors:
                if isinstance(processor_cfg, ActiveServiceConfig):
                    s = self._platform.configuration.service.create_active_service(processor_cfg.type, processor_cfg.owner,
                                                                              processor_cfg.name, r, processor_cfg.access_level,
                                                                              processor_cfg.configuration)  # TODO: Default ID unknown
                else:
                    # TODO: Owner/name/access_level not in the config. It should probably be there. Otherwise we have
                    #       to hardcode.
                    s = self._platform.configuration.service.create_active_service("firewall", "root",
                                                                              "firewall", r, AccessLevel.ELEVATED,
                                                                              {"default_policy": processor_cfg.default_policy}) # TODO: Default ID not known

                    # This is not very pretty. But coding around it would require unnecessary expansion of the API.
                    if isinstance(s.active_service, FirewallImpl):
                        impl: FirewallImpl = s.active_service

                        for chain in processor_cfg.chains:
                            impl.set_default_policy(chain.type, chain.policy)
                            for rule in chain.rules:
                                impl.add_rule(chain.type, rule)

                self._platform.configuration.node.add_traffic_processor(r, s.active_service)

            self._platform.configuration.network.add_node(r)

        # 5) Connections
        for conn_config in self._connections:
            src: NodeImpl = self._obj_refs[self._refs[conn_config.src_ref].id]
            dst: NodeImpl = self._obj_refs[self._refs[conn_config.dst_ref].id]

            conn = self._platform.configuration.network.add_connection(src, dst, conn_config.src_port, conn_config.dst_port)
            conn = ConnectionImpl.cast_from(conn)

            # Propagate new interfaces back into the configuration
            src_config = self._refs[conn_config.src_ref]
            src_port_id = conn.hop.src.port
            src_iface = src.interfaces[src_port_id]

            if conn_config.src_port == -1:
                if isinstance(src_config, RouterConfig):
                    iface_cfg = PortConfig(ip=src_iface.ip, net=src_iface.net, index=src_port_id)
                else:
                    iface_cfg = InterfaceConfig(ip=src_iface.ip, net=src_iface.net, index=src_port_id)

                src_config.interfaces.append(iface_cfg)
                self._refs[iface_cfg.ref] = iface_cfg
                conn_config.src_port = src_port_id

            # Fix interfaces without explicit index (i.e., taken from a list)
            # At this point, ..._config contains only references to other objects, so the complaint about type mismatch
            # is not valid
            self._refs[src_config.interfaces[src_port_id].ref].index = src_port_id

            dst_port_id = conn.hop.dst.port
            dst_iface = dst.interfaces[dst_port_id]
            dst_config = self._refs[conn_config.dst_ref]

            if conn_config.dst_port == -1:
                if isinstance(dst_config, RouterConfig):
                    iface_cfg = PortConfig(ip=dst_iface.ip, net=dst_iface.net, index=dst_port_id)
                else:
                    iface_cfg = InterfaceConfig(ip=dst_iface.ip, net=dst_iface.net, index=dst_port_id)

                dst_config.interfaces.append(iface_cfg.ref)
                self._refs[iface_cfg.ref] = iface_cfg
                conn_config.dst_port = dst_port_id

            # Fix interfaces without explicit index (i.e., taken from a list)
            # At this point, ..._config contains only references to other objects, so the complaint about type mismatch
            # is not valid
            # TODO: I am fixing it to just shut it up. I am not currently sure, what is the reason why only id and not
            #       the config is present here.
            if isinstance(dst_config.interfaces[dst_port_id], str):
                self._refs[dst_config.interfaces[dst_port_id]].index = dst_port_id
            else:
                self._refs[dst_config.interfaces[dst_port_id].ref].index = dst_port_id

        # Sessions
        for session_cfg in self._sessions:
            self._platform.configuration.network.create_session(owner="__system",
                                                                waypoints=session_cfg.waypoints,
                                                                src_service=session_cfg.src_service,
                                                                dst_service=session_cfg.dst_service,
                                                                parent=None,
                                                                defer=True,
                                                                reverse=session_cfg.reverse,
                                                                id=session_cfg.id)



from dataclasses import dataclass, field
from uuid import uuid4
from serde import serialize, coerce

from cyst.api.configuration.configuration import ConfigItem


@serialize(type_check=coerce)
@dataclass
class DataConfig(ConfigItem):
    """ Configuration of a data element.

    This component is currently in a very rudimentary state and the future updates will among other things include the
    option to encrypt the data and add specific flags.

    :param owner: An identity of the owner of the data.
    :type owner: str

    :param path: A path to the data
    :type path: str

    :param description: A textual description of the data. Currently, the only mechanism to distinguish different data.
    :type description: str
    """
    owner: str
    path: str
    description: str
    ref: str = field(default_factory=lambda: str(uuid4()))
    name: str = "__data"
    id: str = ""



# Because we are operating on dataclasses, the initialization order precludes us from having some default initialized
# value, which is a real shame (though understandable)
from copy import copy
from dataclasses import dataclass, field
from enum import Enum, auto
from uuid import uuid4
from typing import Self, Any, List, Union

import serde
from serde import serialize, coerce

import cyst.api.logic.access

"""
Provides an API to configure all aspects of the simulation engine.

Available to: creator, models
Hidden from:  agents
"""


class ConfigItem:
    """
    A superclass of a configuration items.

    Most of the configuration items have an instantiated counterpart produced by the platform. To tie these two object
    classes together, the configuration item has three identification properties: ref, name, and id.

    :param ref: A unique identification of a given configuration item. This parameter is primarily used for referencing
        this configuration from other configurations, usually when a configuration of a particular item is shared among
        multiple other configuration items. Unless there is a good reason to set it explicitly, it is best to let it be
        autogenerated. By default, UUIDs are used.
    :type ref: str

    :param name: An identifier of the object produced from the configuration item. It should be unique among siblings
        within the same parent. As with the ref parameters, it is best to let it be autogenerated, as the system will
        try to come up with a memorable name. In case of name clash among siblings, a suffix is added to the name to
        distinguish different objects. Note that the name is often not accessible through CYST API, except for service
        names that are currently used for message routing and exploit evaluation.
    :type name: str

    :param id: An identifier of the object that is unique within the whole simulation. Unless you have a very good
        reason, you should let it be autogenerated. The id is automatically constructed with the dot notation to
        express the parent-child hierarchy. Note that the id is generally not accessible through CYST API, as it is
        considered an implementation detail that is used for management of simulation components.
    :type id: str
    """
    ref: str
    name: str
    id: str

    # TODO Nested configurations and their copies
    def __call__(self, ref: str | None = None, name: str = "", id: str = "") -> Self:
        """
        Create a copy of a configuration item with a new ref id, name, or id. With this, a configuration item can be
        used as a template to repeat repetitions.

        Example:
            .. code-block:: python

            local_password_auth = AuthenticationProviderConfig(
                provider_type=AuthenticationProviderType.LOCAL,
                token_type=AuthenticationTokenType.PASSWORD,
                token_security=AuthenticationTokenSecurity.SEALED,
                timeout=30
            )

            ssh_provider = local_password_auth(name="openssh_local_pwd_auth")

        :param ref: A new ref id of the configuration item. If none is provided, it is autogenerated. The system does
            not check for duplicate refs, so if you use an already present ref, it will likely end in strange errors
            down the line.
        :type ref: str | None

        :param name: A new name for the resulting object. If none is provided, it is autogenerated.
        :type name: str

        :param id: A new id for the resulting object. It is generally better to not provide an id as it can have
            unintended consequences. It will be autogenerated if none is provided.
        :type id: str

        :return: A copy of a config item.

        """
        new_one = copy(self)
        if ref:
            new_one.ref = ref
        else:
            new_one.ref = str(uuid4())

        if name:
            new_one.name = name

        if id:
            new_one.id = id

        return new_one


@serialize(type_check=coerce)
@dataclass
class ConfigParameter(Any, str):
    """
    ConfigParameter serves as a placeholder, that will be replaced by a concrete value when configuration pass happens.

    Usage example:
        .. code-block:: python

            ...
            PassiveServiceConfig(
                name="bash",
                owner="root",
                version=ConfigParameter("my-custom-version"),
                access_level=ConfigParameter("my-custom-access_level"),
                local=True,
            )
            ...

    :param id: An identifier of the parameter. Must be unique within the configuration.
    :type id: str
    """
    id: str


class ConfigParameterValueType(Enum):
    """
    Configuration parameter can be replaced with a concrete value in the configuration pass in three different ways.
    However, in any case, the developer behind the parametrization is responsible for providing values of the correct
    type. This is not validated, until a configuration pass happens and in case of wrong value type, it will just die
    painfully.

        :VALUE: A specific value is provided.
        :REF: An existing configuration item with a given ref will be used. If multiple config parameters resolve to the
            same ref, then a one instance of the same configuration item will be present in multiple places. This can
            be desired in some cases, but more often it is not. So, beware!
        :REF_COPY: A copy of an existing configuration item will be used, with autogenerated identifiers. Thus, more
            instances of the same type can be created. This is usually the right choice.

    """
    VALUE = auto()
    REF = auto()
    REF_COPY = auto()


@dataclass
class ConfigParameterSingle:
    """
    This represents a single configurable parameter.

    :param name: The name is an identification of a configurable parameter. It must be unique among other configuration
        parameters.
    :type name: str

    :param value_type: The type of configuration parameter, i.e. what type of value will be there after substitution.
    :type value_type: ConfigParameterValueType

    :param description: Additional information about the purpose of the configuration parameter. Most useful for
        user-facing interfaces.
    :type description: str

    :param default: A default value of the parameter. This value will be used if there is no user-provided
        parametrization.
    :type default: Any

    :param parameter_id: An id of a parameter placeholder, which should be replaced by the value of this parameter.
    :type parameter_id: str
    """
    name: str
    value_type: ConfigParameterValueType
    description: str
    default: Any
    parameter_id: str


class ConfigParameterGroupType(Enum):
    """
    When a group of parameters is used, this specifies, how its elements can be selected.

        :ONE: Only one element of the group can be selected.
        :ANY: 0..N elements of the group can be selected.

    """
    ONE = auto()
    ANY = auto()
    # The following are suggested, but ignored until the need arises.
    # AT_LEAST = auto()
    # AT_MOST = auto()
    # ALL = auto()


@dataclass
class ConfigParameterGroupEntry:
    """
    A single choice from the parametrization group. Unlike the single configuration parameter, the group entry directly
    specifies the value. If you require the user to specify something within the group entry, do it two-step like this:

    Example:
        .. code-block:: python

            attacker_service = ActiveServiceConfig(
                type=ConfigParameter("attacker_type"),
                name="attacker",
                owner="attacker",
                access_level=AccessLevel.LIMITED,
                ref="attacker_service_ref"
            )

            ...

            parametrization = ConfigParametrization([
                ConfigParameterGroup(
                    name="Attacker position",
                    group_type=ConfigParameterGroupType.ONE,
                    value_type=ConfigParameterValueType.REF,
                    description="The node, where the attacker starts at.",
                    default="attacker_location_1",
                    options=[
                        ConfigParameterGroupEntry(
                            parameter_id="attacker_location_1",
                            value="attacker_service_ref",
                            description="Node 1"
                        ),
                        ConfigParameterGroupEntry(
                            parameter_id="attacker_location_2",
                            value="attacker_service_ref",
                            description="Node 2"
                        )
                    ]
                )
            ])

    This way, you will enable the user to specify both the position of the attacker from the group and its type.

    :param description: Additional information about the purpose of the configuration parameter. Most useful for
        user-facing interfaces.
    :type description: str

    :param value: The concrete value that will be used if chosen.
    :type value: Any

    :param parameter_id: An id of a parameter placeholder, which should be replaced by the value of this parameter.
    :type parameter_id: str
    """
    description: str
    value: Any
    parameter_id: str


@dataclass
class ConfigParameterGroup:
    """
    Config parameter group represents a choice of zero or more options that a user can make.

    :param name: The name is an identification of the group. It must be unique among other configuration parameters.
    :type name: str

    :param group_type: Specifies a selection method for group entries.
    :type group_type: ConfigParameterGroupType

    :param value_type: Specifies the type of configuration parameter, i.e. what type of value will be there after
        substitution.
    :type value_type: ConfigParameterValueType

    :param description: Provides additional information about the purpose of the group of parameters. Most useful for
        user-facing interfaces.
    :type description: str

    :param default: A `parameter_id` of the entry that is used if the user does not specify anything.
    :type default: str

    :param options: A list of configuration entries to select from.
    :type options: List[ConfigParameterGroupEntry]
    """
    parameter_id: str
    name: str
    group_type: ConfigParameterGroupType
    value_type: ConfigParameterValueType
    description: str
    default: list[str]
    options: List[ConfigParameterGroupEntry]


@dataclass
class ConfigParametrization(ConfigItem):
    """
    A top-level object covering all the parametrization in a given configuration.

    :param parameters: A list of configuration parameters and groups.
    :type parameters: List[Union[ConfigParameterSingle, ConfigParameterGroup]]
    """
    parameters: List[Union[ConfigParameterSingle, ConfigParameterGroup]]
    ref: str = field(default_factory=lambda: str(uuid4()))



import uuid
from abc import ABC

from typing import List, Tuple, Optional, Set
from netaddr import IPAddress

from cyst.api.configuration.logic.access import AccessLevel
from cyst.api.logic.access import Authorization, AuthenticationToken, AuthenticationTokenSecurity, \
    AuthenticationTokenType, AuthenticationProvider, AuthenticationTarget, AuthenticationProviderType, AccessScheme
from cyst.api.logic.data import Data

from cyst.platform.logic.data import DataImpl


class AuthorizationImpl(Authorization):
    def __init__(self, identity: str = "", nodes: List[str] = None, services: List[str] = None,
                 access_level: AccessLevel = AccessLevel.NONE, id: Optional[str] = None, token: Optional[str] = None):
        if services is None or not services:
            services = ["*"]
        if nodes is None or not nodes:
            nodes = ["*"]
        self._id = id
        self._identity = identity
        self._nodes = nodes
        self._services = services
        self._access_level = access_level
        self._token = token
        self._expiration = -1  # TODO

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, AuthorizationImpl):
            return NotImplemented
        if not other:
            return False

        other = AuthorizationImpl.cast_from(other)
        return self.id == other.id or (
                self.identity == other.identity and
                self.nodes == other.nodes and
                self.services == other.services and
                self.access_level == other.access_level and
                self.token == other.token
        )

    @property
    def id(self) -> str:
        #MYPY: complains about optionality of the id, it is also noted, that it is suspicious in Authorization class
        return self._id

    @id.setter
    def id(self, value: str) -> None:
        self._id = value

    @property
    def identity(self) -> str:
        return self._identity

    @identity.setter
    def identity(self, value: str) -> None:
        self._identity = value

    @property
    def nodes(self) -> List[str]:
        return self._nodes

    @nodes.setter
    def nodes(self, value: List[str]) -> None:
        self._nodes = value

    @property
    def services(self) -> List[str]:
        return self._services

    @services.setter
    def services(self, value: List[str]) -> None:
        self._services = value

    @property
    def access_level(self) -> AccessLevel:
        return self._access_level

    @access_level.setter
    def access_level(self, value: AccessLevel) -> None:
        self._access_level = value

    @property
    def token(self) -> Optional[uuid.UUID]:
        return self._token #MYPY: Authorization has token as not null, however there are calls to this implementation, that do not set it and therefore it can be null. Is relaxing authorization by making token null the correct choice?
    #Also UUID and str is not unified


    @token.setter
    def token(self, value: uuid.UUID) -> None:
        self._token = value #MYPY: uid vs str

    def matching_id(self, other: Authorization):
        return self.id == AuthorizationImpl.cast_from(other).id

    def __str__(self) -> str:
        return "[Id: {}, Identity: {}, Nodes: {}, Services: {}, Access Level: {}, Token: {}]".format(self.id,
                                                                                                     self.identity,
                                                                                                     self.nodes,
                                                                                                     self.services,
                                                                                                     self.access_level.name,
                                                                                                     self.token)

    def __repr__(self) -> str:
        return self.__str__()

    @staticmethod
    def cast_from(o: Authorization) -> 'AuthorizationImpl':
        if isinstance(o, AuthorizationImpl):
            return o
        else:
            raise ValueError("Malformed underlying object passed with the Authorization interface")

    @property
    def expiration(self) -> int:
        return self._expiration


class PolicyStats:
    def __init__(self, authorization_entry_count: int = 0):
        self._authorization_entry_count = authorization_entry_count

    @property
    def authorization_entry_count(self):
        return self._authorization_entry_count


# ----------------------------------------------------------------------------------------------------------------------
# New version
# ----------------------------------------------------------------------------------------------------------------------

class AuthenticationTokenImpl(AuthenticationToken):

    def __init__(self, token_type: AuthenticationTokenType, security: AuthenticationTokenSecurity, identity: str,
                 is_local: bool):
        self._type = token_type
        self._security = security
        self._identity = identity
        self._is_local = is_local

        # create data according to the security
        # TODO: Until the concept of sealed data is introduced in the code, all is assumed to be OPEN
        value = uuid.uuid4()
        self._content = None

    @property
    def type(self) -> AuthenticationTokenType:
        return self._type

    @property
    def is_local(self):
        return self._is_local

    @property
    def security(self) -> AuthenticationTokenSecurity:
        return self._security

    @property
    def identity(self) -> str:
        return self._identity #MYPY: Setter is defined in AuthenticationToken, is it OK to add here? It should be here as well

    def copy(self) -> Optional[AuthenticationToken]:
        pass # TODO different uuid needed????

    @property
    def content(self) -> Optional[Data]:
        return self._content

    @staticmethod
    def is_local_instance(obj: AuthenticationToken):
        if isinstance(obj, AuthenticationTokenImpl):
            return obj.is_local
        return False

    def _set_content(self, value) -> 'AuthenticationToken':
        # this is only needed until we resolve hashed/encrypted data
        self._content = DataImpl("", value)
        return self  # just so we can chain call with constructor



class AuthenticationTargetImpl(AuthenticationTarget):
    def __init__(self, tokens: List[AuthenticationTokenType], service: Optional[str] = None,
                 ip: Optional[IPAddress] = None):
        self._address = ip
        self._service = service
        self._tokens = tokens

    @property
    def address(self) -> Optional[IPAddress]:
        return self._address

    @address.setter
    def address(self, ip: IPAddress):
        self._address = ip

    @property
    def service(self) -> str:
        return self._service #MYPY: Service is defined as not optional in AuthenticationTarget, but constructor of impl has it optional and is called in that way

    @service.setter
    def service(self, serv: str):
        self._service = serv

    @property
    def tokens(self) -> List[AuthenticationTokenType]:
        return self._tokens



class AccessSchemeImpl(AccessScheme):
    def __init__(self, id: str):
        self._providers = []
        self._authorizations: List[Authorization] = []
        self._identities: List[str] = []
        self._id = id

    def add_provider(self, provider: AuthenticationProvider):
        self._providers.append((provider, len(self._providers)))

    def add_identity(self, identity: str):
        self._identities.append(identity)

    def add_authorization(self, auth: Authorization):
        self._authorizations.append(auth)

    def remove_authorization(self, auth: Authorization):
        # removing authorization means setting access level to NONE
        auth.access_level = AccessLevel.NONE

    def register_authorization(self, auth: Authorization) -> None:
        identity = AuthorizationImpl.cast_from(auth).identity
        if identity not in self.identities:
            self.add_identity(identity)
        self.add_authorization(auth)

    def is_local(self) -> bool:
        return all(p.type is AuthenticationProviderType.LOCAL for p, _ in self.factors)

    def match_tokens(self, tokens: List[AuthenticationToken]) -> bool:
        return len(tokens) == len(self.factors) and \
                all(AuthenticationProviderImpl.cast_from(factor).is_token_matching(token)
                    for (factor, _), token in zip(self.factors, tokens))

    def create_access(self, identity: str, access_level: AccessLevel, tokens: List[AuthenticationToken]) -> List[AuthenticationToken]:
        self.register_authorization(AuthorizationImpl(identity, id=str(uuid.uuid4()), access_level=access_level))

        # no tokens supplied, create and return new ones
        if not tokens:
            return [AuthenticationProviderImpl.cast_from(p).register_new_token(identity) for p, _ in self.factors]

        # otherwise add tokens to each factor one by one
        for (factor, _), token in zip(self.factors, tokens):
            AuthenticationProviderImpl.cast_from(factor).add_token(token)

        return tokens

    def modify_access(self, identity: str, access_level: AccessLevel) -> None:
        # WARN: this will raise exception if identity not present
        template = next(a for a in self.authorizations if a.identity == identity)
        AuthorizationImpl.cast_from(template).access_level = access_level

    @property
    def factors(self) -> List[Tuple[AuthenticationProvider, int]]:
        return self._providers
    # TODO : what is the number?? I will just use order ATM

    @property
    def identities(self) -> List[str]:
        return self._identities

    @property
    def authorizations(self) -> List[Authorization]:
        return self._authorizations

    @property
    def id(self) -> str:
        return self._id

    @staticmethod
    def cast_from(other: AccessScheme):
        if isinstance(other, AccessSchemeImpl):
            return other
        else:
            raise ValueError("Malformed underlying object passed with the AccessScheme interface")


class AuthenticationProviderImpl(AuthenticationProvider):

    class AuthenticationTokenState(ABC):
        def __init__(self, token: AuthenticationToken):
            self._token = token
            self._enabled = True
            self._disabled_until = 0

        @property
        def token(self) -> AuthenticationToken:
            return self._token

        @property
        def enabled(self) -> bool:
            return self._enabled

        def enable_token(self) -> None:
            self._enabled = True
            self._disabled_until = 0

        def disable_token(self, time: int) -> None:
            self._enabled = False
            self._disabled_until = time

        def is_token_valid(self, time: int) -> bool:
            if self._disabled_until < time:
                # if _disabled_until expired, token is active at the time 'time', therefore enable token
                self.enable_token()
            return self._enabled

    def __init__(self, provider_type: AuthenticationProviderType, token_type: AuthenticationTokenType,
                 security: AuthenticationTokenSecurity, ip: Optional[IPAddress], timeout: int, id: str):

        self._provider_type = provider_type
        self._token_type = token_type
        self._security = security
        self._timeout = timeout

        self._tokens: Set[AuthenticationToken] = set()
        self._target = self._create_target()

        self._id = id

        if provider_type != AuthenticationProviderType.LOCAL and ip is None:
            raise RuntimeError("Non-local provider needs ip address")
        self._set_address(ip)

    @property
    def type(self) -> AuthenticationProviderType:
        return self._provider_type

    @property
    def target(self) -> AuthenticationTarget:
        return self._target

    @property
    def token_type(self):
        return self._token_type

    @property
    def security(self):
        return self._security

    @property
    def id(self) -> str:
        return self._id

    def token_is_registered(self, token: AuthenticationToken):
        for token_state in set(self._tokens):
            if token_state.token.identity == token.identity and token.content is not None:
                # This is pretty weak but until encrypted/hashed stuff is implemented its okay for testing
                return True
        return False

    def create_token(self, identity: str) -> AuthenticationToken:
        return AuthenticationTokenImpl(self._token_type, self._security, identity,
                self._provider_type is AuthenticationProviderType.LOCAL)._set_content(uuid.uuid4())

    def register_new_token(self, identity: str) -> AuthenticationToken:
        token = self.create_token(identity)
        self.add_token(token)
        return token

    def is_token_matching(self, token: AuthenticationToken) -> bool:
        return isinstance(token, AuthenticationTokenImpl) and \
            token.security is self.security and \
            token.type is self.token_type and \
            token.is_local == (self.type is AuthenticationProviderType.LOCAL)

    def add_token(self, token: AuthenticationToken):
        self._tokens.add(self.AuthenticationTokenState(token))

    def disable_token(self, token: AuthenticationToken, time: int):
        for token_state in set(self._tokens):
            if token_state.token == token:
                token_state.disable_token(time)

    def enable_token(self, token: AuthenticationToken):
        for token_state in set(self._tokens):
            if token_state.token == token:
                token_state.enable_token()

    def is_token_valid(self, token: AuthenticationToken, time: int) -> bool:
        for token_state in set(self._tokens):
            # TODO: the token is matched only by identity on the premise that one authentication factor has only one
            #       auth token for identity. Evaluate if this is really true or not
            if token_state.token.identity == token.identity:
                return token_state.is_token_valid(time)

    def remove_token_by_identity(self, identity: str):
        for token_state in set(self._tokens):
            if token_state.token.identity == identity:
                self._tokens.remove(token_state)

    def get_token_by_identity(self, identity: str):
        for token_state in set(self._tokens):
            if token_state.token.identity == identity:
                return token_state.token

    def _create_target(self):
        # TODO: inherit from provider? or should we do something else?
        return AuthenticationTargetImpl([self._token_type])

    def set_service(self, srv_id: str):

        if self._target.service is None:
            self._target.service = srv_id
        # Due to the way a configuration works from version 0.6.0, the same service can be set multiple times. So we
        # just ignore it.
        elif self._target.service == srv_id:
            pass
        else:
            raise RuntimeError  # TODO check what should be done here, exception might be too harsh

    def _set_address(self, ip: IPAddress):
        if self._target.address is None:
            self._target.address = ip
        else:
            raise RuntimeError

    @staticmethod
    def cast_from(other: AuthenticationProvider):
        if isinstance(other, AuthenticationProviderImpl):
            return other
        else:
            raise ValueError("Malformed underlying object passed with the AuthenticationProvider interface")



from abc import ABC, abstractmethod
from collections.abc import Sequence
from dataclasses import dataclass
from deprecated.sphinx import versionchanged, versionadded
from enum import Enum
from typing import List, Tuple, Optional, Any, Dict, Union

from cyst.api.logic.exploit import Exploit
from cyst.api.environment.platform_specification import PlatformSpecification


class ActionParameterDomainType(Enum):
    """
    Specifies a type of the action parameter domain.

    Possible values:
        :ANY: The parameters can take any values. If possible avoid this domain, as it does not enable sampling of
            values, thus being less useful for automated processing and learning.
        :RANGE: The parameters belong to a certain numeric range, e.g., port numbers.
        :OPTIONS: The parameters are a collection of values to choose from.
    """
    ANY = 0,
    RANGE = 1,
    OPTIONS = 2


class ActionParameterDomain(Sequence):
    """
    Specifies a domain of action parameters.
    """
    @property
    @abstractmethod
    def type(self) -> ActionParameterDomainType:
        """
        Returns a type of the action parameter domain.

        :rtype: ActionParameterDomainType
        """

    @property
    @abstractmethod
    def range_min(self) -> int:
        """
        Returns a minimal value of the range, if the domain is of RANGE type. Otherwise the value is undefined.

        :rtype: int
        """

    @property
    @abstractmethod
    def range_max(self) -> int:
        """
        Returns a maximal value of the range, if the domain is of RANGE type. Otherwise the value is undefined.

        :rtype: int
        """

    @property
    @abstractmethod
    def range_step(self) -> int:
        """
        Returns a gap between two values in a range. E.g., a range [1,3,5,7] has a step = 2.

        :rtype: int
        """

    @property
    @abstractmethod
    def options(self) -> List[Any]:
        """
        Returns a list of possible values in the domain, if the type of the domain is OPTIONS. Otherwise and empty list.

        :rtype: List[Any]
        """

    @abstractmethod
    def validate(self, value: Any) -> bool:
        """
        Checks, whether a value belongs into a domain.

        :param value: A value to check.
        :type value: Any
        """

    @property
    @abstractmethod
    def default(self) -> Any:
        """
        :return: A value from the domain that is considered a default one.
        """

    @abstractmethod
    def __getitem__(self, item: int) -> Any:
        """
        This method is here for the domain to be treated as a range, so it can be, e.g., randomly sampled. It is not
        intended to be used directly. This does not work with domains of type ANY.

        :param item: An index of the item to get.
        :type item: int

        :return: The item with given index.
        """

    @abstractmethod
    def __len__(self) -> int:
        """
        This method is here for the domain to be treated as a range, so it can be, e.g., randomly sampled. It is not
        intended to be used directly. This does not work with domains of type ANY.

        :return: The number of items in the domain.
        """


class ActionParameterType(Enum):
    """
    Specifies a type of action parameters.

    Possible values:
        :NONE: Type not specified, or outside the current domain of parameter types. This should be used only after
            due consideration, as it is better to expand this domain than to use NONE. It may interfere with the
            learning process.
        :IDENTITY: The parameter is an identity of a user in the system.
        :IDENTIFIER: The parameter is some kind of identifier.
        :DURATION: The parameter represents a time duration in simulated time units.
        :TOKEN: The parameter is an instance of a token object, such as cyst.api.logic.access.AuthenticationToken.
        :DATA: Added as a temporary solution for actions and communication that carries actual data not encoded into
            the message structure. Given the complicated definition of parameters and factual uselessness of fitting
            data into a domain that can be sampled, this will likely be replaced by a more top-level solution.
    """
    NONE = 0,
    IDENTITY = 1,
    IDENTIFIER = 2,
    DURATION = 3,
    TOKEN = 4,
    DATA = 5


@dataclass
class ActionParameter:
    """
    Action parameter represents a mechanism to further specify details of actions. An action can have an arbitrary
    number of parameters.

    :param type: The type of the parameter.
    :type type: ActionParameterType

    :param name: A name of the parameter, which is unique within an action.
    :type name: str

    :param domain: A domain of the parameter.
    :type domain: ActionParameterDomain

    :param value: A default value of the parameter.
    :type value: Optional[Any]
    """
    type: ActionParameterType
    name: str
    domain: ActionParameterDomain
    value: Optional[Any] = None


@versionadded(version="0.6.0")
class ActionType(Enum):
    """
    Specifies the type of the action. The type governs, how the action is being evaluated during a simulation.

    Possible values:
        :COMPONENT: Component actions can only be expressed within the simulation as a part of another actions. They
            do not have their own semantics expressed through a behavioral model. They are primarily used to provide
            parametrized description of action elements, such as network flows. The intended recipient for these type
            of actions are metadata providers, but behavioral models can also use them to fine-tune evaluation of
            higher-level actions.
        :DIRECT: Direct actions are expressed through a behavioral model. A direct action is always evaluated in the
            context of the target it reached.
        :COMPOSITE: Composite actions enable specification of higher-order actions. A composite action is defined by a
            flow of other actions - composite or direct. Semantics of composite actions are defined through its
            constituent actions, which get processed and evaluated in a given flow. A composite action is evaluated in
            the context of its source.
    """
    COMPONENT = 0
    DIRECT = 1
    COMPOSITE = 2


@versionchanged(version="0.6.0", reason="Removed action tokens until a new action meta-model is polished. Added type and platform specification.")
@dataclass
class ActionDescription:
    """
    A description of an action in the form that is used for registering into the system.

    :param id: A name of the action. The combination of name and execution environment must be unique across all
        behavioral models, so it is advisable to use own namespace for each model.
    :type id: str

    :param type: A type of the action.
    :type type: ActionType

    :param description: A textual description of the action.
    :type description: str

    :param parameters: A list of action parameters.
    :type parameters: List[ActionParameter]

    :param platform: The platform, where the action is valid. One or more platforms can be specified.
        If none is provided, the default CYST simulation is assumed.
    :type platform: Union[None, PlatformSpecification, List[PlatformSpecification]]
    """
    id: str
    type: ActionType
    description: str
    parameters: List[ActionParameter]
    platform: Union[None, PlatformSpecification, List[PlatformSpecification]] = None


class Action(ABC):
    """
    An action represent an activity that an actor within simulation can do. Usage of actions is mostly the domain of
    active services. An action does not carry any semantics by itself. The semantics is provided by interpreters /
    behavioral models. One action can therefore have different impact depending on the model chosen.
    """
    @property
    @abstractmethod
    def id(self) -> str:
        """
        A unique identifier of the action within simulation. The identifier is expected to be a serialization of
        a position in hierarchy, separated by a colon. E.g., "aif:privilege_escalation:root_privilege_escalation".

        :rtype: str
        """

    @property
    @abstractmethod
    def description(self) -> str:
        """
        A textual description of the action.

        :rtype: str
        """

    @property
    @abstractmethod
    def fragments(self) -> List[str]:
        """
        The list of fragments of the id (i.e., what is left after removing the colons). E.g, ["aif",
        "privilege_escalation", "root_privilege escalation"].

        :rtype: List[str]
        """

    @property
    @abstractmethod
    def namespace(self) -> str:
        """
        Currently a shorthand for fragments()[0], but that can change in the future, as the namespace can span multiple
        fragments.

        :rtype: str
        """

    @property
    @abstractmethod
    def exploit(self) -> Optional[Exploit]:
        """
        Returns an exploit currently associated with the action.

        :rtype: Optional[Exploit]
        """

    @abstractmethod
    def set_exploit(self, exploit: Optional[Exploit]) -> None:
        """
        Sets an exploit associated with the action. TODO: Why the hell is this not a setter of the exploit property?

        :param exploit: The exploit to use. If None is passed as a value, removes the exploit from the action.
        :type exploit: Optional[Exploit]
        """

    @property
    @abstractmethod
    def parameters(self) -> Dict[str, ActionParameter]:
        """
        Returns a collection of action parameters.

        :rtype: Dict[str, ActionParameter]
        """

    @abstractmethod
    def add_parameters(self, *params: ActionParameter) -> None:
        """
        Adds one or more parameters to the action. If a parameter of the same name is present, it gets overwritten.

        :param params: One or more action parameters.
        :type params: ActionParameter
        """

    @versionadded(version="0.6.0")
    @property
    @abstractmethod
    def components(self) -> List['Action']:
        """
        Returns a list of actions that constitute components of this action.

        :rtype: List[Action]
        """
        pass

    @versionadded(version="0.6.0")
    @abstractmethod
    def add_components(self, *components: 'Action') -> None:
        """
        Adds one or more component actions to the actions. The system validates, whether an action has ActionType equal
        to COMPONENT. If not, an exception is thrown.

        :param components: One or more component Actions.
        :type components: Action
        """

    @abstractmethod
    def copy(self) -> 'Action':
        """
        Returns a copy of the action. This is useful for keeping copies of action with different parameters or exploits.

        :rtype: Action
        """



from dataclasses import dataclass
from enum import Enum, auto


class PlatformType(Enum):
    """
    Differentiates the modus operandi of the platform. As there is a blurry line between the simulation and emulation,
    we differentiate them in respect their control of time.

    Possible values:

        :SIMULATED_TIME: A platform fully controls execution time.
        :REAL_TIME: A platform is operating in real-time.
    """
    SIMULATED_TIME = auto()
    REAL_TIME = auto()


@dataclass(frozen=True)
class PlatformSpecification:
    """
    A specification of a platform. Each specification must be unique among the loaded modules.

    :param type: Type of execution of the platform.
    :type type: PlatformType

    :param provider: A unique identifier of the platform. This string is used by the users to select the platform in the
        call to Environment's :func:`create` method.
    :type provider: str
    """
    type: PlatformType
    provider: str



import logging

from typing import Any, Dict, Optional, Tuple

from cyst.api.environment.message import Message, MessageType, Request, Status, StatusOrigin, StatusValue
from cyst.api.environment.messaging import EnvironmentMessaging
from cyst.api.environment.resources import EnvironmentResources
from cyst.api.host.service import ActiveService, ActiveServiceDescription


class ForwardShell(ActiveService):

    def __init__(self,
                 msg: EnvironmentMessaging,
                 res: EnvironmentResources,
                 id: str,
                 args: Optional[Dict[str, Any]] = None) -> None:
        self._messaging = msg
        self._resources = res
        self._id = id
        self._log = logging.getLogger("services.forward_shell")
        self._ignore_requests: bool = args is None or args.get("ignore_requests", True)

        self._success_sent = False
        self._origin_request = None
        self._session = None

    async def run(self) -> None:
        self._log.info("Launched a forward shell service")

    async def process_message(self, message: Message) -> Tuple[bool, int]:
        self._log.debug(f"Processing message {message.id} : {message}")

        if message.type is not MessageType.REQUEST:
            return False, 0

        request = message.cast_to(Request)

        if request.action.id == "cyst:active_service:open_session":
            self._log.debug(f"Openning session for {request.src_service}")
            self._origin_request = request
            self._respond_with_session(request)

            if not self._success_sent:
                self._send_success_to_origin()
                self._success_sent = True

            return True, 1

        if not self._ignore_requests:
            self._respond_with_error(request, f"Invalid action {request.action.id}")

        return False, 1

    def _respond_with_session(self, request: Request) -> None:
        self._session = self._messaging.open_session(request)
        response = self._messaging.create_response(request,
                                                   Status(StatusOrigin.SERVICE,
                                                          StatusValue.SUCCESS),
                                                   session=self._session)
        self._messaging.send_message(response)

    def _respond_with_error(self, request: Request, error: str) -> None:
        response = self._messaging.create_response(
            request, Status(StatusOrigin.SERVICE, StatusValue.FAILURE), error)

        self._messaging.send_message(response)

    def _send_success_to_origin(self) -> None:
        request = self._origin_request
        response = self._messaging.create_response(request,
                                                   Status(StatusOrigin.SERVICE,
                                                          StatusValue.SUCCESS),
                                                   f"Session opened as a reaction to {self._origin_request.id}",
                                                   session=self._session)
        self._messaging.send_message(response)
        self._success_sent = True


def create_shell(msg: EnvironmentMessaging, res: EnvironmentResources,
                 id:str, args: Optional[Dict[str, Any]]) -> ActiveService:
    return ForwardShell(msg, res, id, args)


service_description = ActiveServiceDescription(
    "forward_shell",
    "A service acting as a forward shell. It will create a session given the correct action",
    create_shell)



from abc import ABC, abstractmethod

from cyst.api.environment.configuration import RuntimeConfiguration
from cyst.api.environment.stats import Statistics
from cyst.api.environment.stores import ServiceStore, DataStore


class EnvironmentInfrastructure(ABC):
    """
    This interface provides access to environment resources that are aimed at behavioral models and to applications
    utilizing CYST.
    """

    @property
    @abstractmethod
    def statistics(self) -> Statistics:
        """
        Statistics track basic information about the simulation runs.

        :rtype: Statistics
        """

    @property
    @abstractmethod
    def data_store(self) -> DataStore:
        """
        Data store provides access to storing run-related data for later analysis.

        :rtype: DataStore
        """

    @property
    @abstractmethod
    def service_store(self) -> ServiceStore:
        """
        Provides access to creation and querying of active services.

        :rtype: ServiceStore
        """

    @property
    @abstractmethod
    def runtime_configuration(self) -> RuntimeConfiguration:
        """
        Provides access to parameters that were set either through a command line or through environmental variables.

        :rtype: RuntimeConfiguration
        """



from abc import ABC, abstractmethod

from cyst.api.environment.external import ExternalResources
from cyst.api.environment.stores import ActionStore, ExploitStore
from cyst.api.environment.clock import Clock


class EnvironmentResources(ABC):
    """
    This interface provides access to resources that can be used by services within the simulation.
    """

    @property
    @abstractmethod
    def action_store(self) -> ActionStore:
        """
        Through action store, services access actions that are available to them in the simulation.

        :rtype: ActionStore
        """

    @property
    @abstractmethod
    def exploit_store(self) -> ExploitStore:
        """
        Through exploit store, services access exploits that can be tied to actions.

        :rtype: ExploitStore
        """

    @property
    @abstractmethod
    def clock(self) -> Clock:
        """
        Clock provides a mean to track simulation and hybrid time.

        :rtype: Clock
        """

    @property
    @abstractmethod
    def external(self) -> ExternalResources:
        """
        External resources enable to interface simulation with data and services outside.

        :rtype: ExternalResources
        """